# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Paired text data that consists of source text and target text.
"""
import math
from typing import List, Optional, Tuple

import torch

from texar.torch.data.data.data_base import (
    DataSource, FilterDataSource, ZipDataSource)
from texar.torch.data.data.dataset_utils import Batch, padded_batch
from texar.torch.data.data.mono_text_data import (
    MonoTextData, _LengthFilterMode, _default_mono_text_dataset_hparams)
from texar.torch.data.data.text_data_base import (
    TextDataBase, TextLineDataSource)
from texar.torch.data.embedding import Embedding
from texar.torch.data.vocabulary import SpecialTokens, Vocab
from texar.torch.hyperparams import HParams
from texar.torch.utils import utils

__all__ = [
    "_default_paired_text_dataset_hparams",
    "PairedTextData",
]


def _default_paired_text_dataset_hparams():
    r"""Returns hyperparameters of a paired text dataset with default values.

    See :meth:`texar.torch.data.PairedTextData.default_hparams` for details.
    """
    source_hparams = _default_mono_text_dataset_hparams()
    source_hparams["bos_token"] = None
    source_hparams["data_name"] = "source"
    target_hparams = _default_mono_text_dataset_hparams()
    target_hparams.update(
        {
            "vocab_share": False,
            "embedding_init_share": False,
            "processing_share": False,
            "data_name": "target"
        }
    )
    return {
        "source_dataset": source_hparams,
        "target_dataset": target_hparams
    }


class PairedTextData(TextDataBase[Tuple[List[str], List[str]],
                                  Tuple[List[str], List[str]]]):
    r"""Text data processor that reads parallel source and target text.
    This can be used in, e.g., seq2seq models.

    Args:
        hparams (dict): Hyperparameters. See :meth:`default_hparams` for the
            defaults.
        device: The device of the produced batches. For GPU training, set to
            current CUDA device.

    By default, the processor reads raw data files, performs tokenization,
    batching and other pre-processing steps, and results in a Dataset
    whose element is a python `dict` including six fields:

    "source_text":
        A list of ``[batch_size]`` elements each containing a list of
        **raw** text tokens of source sequences. Short sequences in the
        batch are padded with **empty string**. By default only ``EOS``
        token is appended to each sequence. Out-of-vocabulary tokens are
        **NOT** replaced with ``UNK``.
    "source_text_ids":
        A list of ``[batch_size]`` elements each containing a list of token
        indexes of source sequences in the batch.
    "source_length":
        A list of ``[batch_size]`` elements of integers containing the length
        of each source sequence in the batch.
    "target_text":
        A list same as "source_text" but for target sequences. By default
        both BOS and EOS are added.
    "target_text_ids":
        A list same as "source_text_ids" but for target sequences.
    "target_length":
        An list same as "source_length" but for target sequences.

    The above field names can be accessed through :attr:`source_text_name`,
    :attr:`source_text_id_name`, :attr:`source_length_name`, and those prefixed
    with ``target_``, respectively.

    Example:

    .. code-block:: python

        hparams={
            'source_dataset': {'files': 's', 'vocab_file': 'vs'},
            'target_dataset': {'files': ['t1', 't2'], 'vocab_file': 'vt'},
            'batch_size': 1
        }
        data = PairedTextData(hparams)
        iterator = DataIterator(data)

        for batch in iterator:
            # batch contains the following
            # batch_ == {
            #    'source_text': [['source', 'sequence', '<EOS>']],
            #    'source_text_ids': [[5, 10, 2]],
            #    'source_length': [3]
            #    'target_text': [['<BOS>', 'target', 'sequence', '1',
                                '<EOS>']],
            #    'target_text_ids': [[1, 6, 10, 20, 2]],
            #    'target_length': [5]
            # }

    """

    def __init__(self, hparams, device: Optional[torch.device] = None):
        self._hparams = HParams(hparams, self.default_hparams())

        src_hparams = self.hparams.source_dataset
        tgt_hparams = self.hparams.target_dataset

        # create vocabulary
        self._src_bos_token = src_hparams["bos_token"]
        self._src_eos_token = src_hparams["eos_token"]
        self._src_transforms = src_hparams["other_transformations"]
        self._src_vocab = Vocab(src_hparams.vocab_file,
                                bos_token=src_hparams.bos_token,
                                eos_token=src_hparams.eos_token)

        if tgt_hparams["processing_share"]:
            self._tgt_bos_token = src_hparams["bos_token"]
            self._tgt_eos_token = src_hparams["eos_token"]
        else:
            self._tgt_bos_token = tgt_hparams["bos_token"]
            self._tgt_eos_token = tgt_hparams["eos_token"]
        tgt_bos_token = utils.default_str(self._tgt_bos_token,
                                          SpecialTokens.BOS)
        tgt_eos_token = utils.default_str(self._tgt_eos_token,
                                          SpecialTokens.EOS)
        if tgt_hparams["vocab_share"]:
            if tgt_bos_token == self._src_vocab.bos_token and \
                    tgt_eos_token == self._src_vocab.eos_token:
                self._tgt_vocab = self._src_vocab
            else:
                self._tgt_vocab = Vocab(src_hparams["vocab_file"],
                                        bos_token=tgt_bos_token,
                                        eos_token=tgt_eos_token)
        else:
            self._tgt_vocab = Vocab(tgt_hparams["vocab_file"],
                                    bos_token=tgt_bos_token,
                                    eos_token=tgt_eos_token)

        # create embeddings
        self._src_embedding = MonoTextData.make_embedding(
            src_hparams.embedding_init, self._src_vocab.token_to_id_map_py)

        if self._hparams.target_dataset.embedding_init_share:
            self._tgt_embedding = self._src_embedding
        else:
            tgt_emb_file = tgt_hparams.embedding_init["file"]
            self._tgt_embedding = None
            if tgt_emb_file is not None and tgt_emb_file != "":
                self._tgt_embedding = MonoTextData.make_embedding(
                    self._tgt_vocab.token_to_id_map_py,
                    tgt_hparams.embedding_init)

        # create data source
        self._src_delimiter = src_hparams.delimiter
        self._src_max_seq_length = src_hparams.max_seq_length
        self._src_length_filter_mode = _LengthFilterMode(
            src_hparams.length_filter_mode)
        self._src_pad_length = self._src_max_seq_length
        if self._src_pad_length is not None:
            self._src_pad_length += sum(int(x is not None and x != '')
                                        for x in [src_hparams.bos_token,
                                                  src_hparams.eos_token])

        src_data_source = TextLineDataSource(
            src_hparams.files, compression_type=src_hparams.compression_type)

        self._tgt_transforms = tgt_hparams["other_transformations"]
        self._tgt_delimiter = tgt_hparams.delimiter
        self._tgt_max_seq_length = tgt_hparams.max_seq_length
        self._tgt_length_filter_mode = _LengthFilterMode(
            tgt_hparams.length_filter_mode)
        self._tgt_pad_length = self._tgt_max_seq_length
        if self._tgt_pad_length is not None:
            self._tgt_pad_length += sum(int(x is not None and x != '')
                                        for x in [tgt_hparams.bos_token,
                                                  tgt_hparams.eos_token])

        tgt_data_source = TextLineDataSource(
            tgt_hparams.files, compression_type=tgt_hparams.compression_type)

        data_source: DataSource[Tuple[List[str], List[str]]]
        data_source = ZipDataSource(  # type: ignore
            src_data_source, tgt_data_source)
        if ((self._src_length_filter_mode is _LengthFilterMode.DISCARD and
             self._src_max_seq_length is not None) or
                (self._tgt_length_filter_mode is _LengthFilterMode.DISCARD and
                 self._tgt_length_filter_mode is not None)):
            max_source_length = self._src_max_seq_length or math.inf
            max_tgt_length = self._tgt_max_seq_length or math.inf

            def filter_fn(raw_example):
                return (len(raw_example[0]) <= max_source_length and
                        len(raw_example[1]) <= max_tgt_length)

            data_source = FilterDataSource(data_source, filter_fn)

        super().__init__(data_source, hparams, device=device)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of default hyperparameters.

        .. code-block:: python

            {
                # (1) Hyperparams specific to text dataset
                "source_dataset": {
                    "files": [],
                    "compression_type": None,
                    "vocab_file": "",
                    "embedding_init": {},
                    "delimiter": None,
                    "max_seq_length": None,
                    "length_filter_mode": "truncate",
                    "pad_to_max_seq_length": False,
                    "bos_token": None,
                    "eos_token": "<EOS>",
                    "other_transformations": [],
                    "variable_utterance": False,
                    "utterance_delimiter": "|||",
                    "max_utterance_cnt": 5,
                    "data_name": "source",
                },
                "target_dataset": {
                    # ...
                    # Same fields are allowed as in "source_dataset" with the
                    # same default values, except the
                    # following new fields/values:
                    "bos_token": "<BOS>"
                    "vocab_share": False,
                    "embedding_init_share": False,
                    "processing_share": False,
                    "data_name": "target"
                }
                # (2) General hyperparams
                "num_epochs": 1,
                "batch_size": 64,
                "allow_smaller_final_batch": True,
                "shuffle": True,
                "shuffle_buffer_size": None,
                "shard_and_shuffle": False,
                "num_parallel_calls": 1,
                "prefetch_buffer_size": 0,
                "max_dataset_size": -1,
                "seed": None,
                "name": "paired_text_data",
                # (3) Bucketing
                "bucket_boundaries": [],
                "bucket_batch_sizes": None,
                "bucket_length_fn": None,
            }

        Here:

        1. Hyperparameters in the :attr:`"source_dataset"` and
           :attr:`"target_dataset"` fields have the same definition as those
           in :meth:`texar.torch.data.MonoTextData.default_hparams`, for source
           and target text, respectively.

           For the new hyperparameters in :attr:`"target_dataset"`:

           `"vocab_share"`: bool
               Whether to share the vocabulary of source.
               If `True`, the vocab file of target is ignored.

           `"embedding_init_share"`: bool
               Whether to share the embedding initial value of source. If
               `True`, :attr:`"embedding_init"` of target is ignored.

              :attr:`"vocab_share"` must be true to share the embedding
              initial value.

           `"processing_share"`: bool
               Whether to share the processing configurations of source,
               including `"delimiter"`, `"bos_token"`, `"eos_token"`, and
               `"other_transformations"`.

        2. For the **general** hyperparameters, see
           :meth:`texar.torch.data.DatasetBase.default_hparams` for details.

        3. For **bucketing** hyperparameters, see
           :meth:`texar.torch.data.MonoTextData.default_hparams` for details,
           except that the default `"bucket_length_fn"` is the maximum sequence
           length of source and target sequences.

           .. warning::
               Bucketing is not yet supported. These options will be ignored.

        """
        hparams = TextDataBase.default_hparams()
        hparams["name"] = "paired_text_data"
        hparams.update(_default_paired_text_dataset_hparams())
        return hparams

    @staticmethod
    def make_embedding(src_emb_hparams, src_token_to_id_map,
                       tgt_emb_hparams=None, tgt_token_to_id_map=None,
                       emb_init_share=False):
        r"""Optionally loads source and target embeddings from files (if
        provided), and returns respective :class:`texar.torch.data.Embedding`
        instances.
        """
        src_embedding = MonoTextData.make_embedding(
            src_emb_hparams, src_token_to_id_map)

        if emb_init_share:
            tgt_embedding = src_embedding
        else:
            tgt_emb_file = tgt_emb_hparams["file"]
            tgt_embedding = None
            if tgt_emb_file is not None and tgt_emb_file != "":
                tgt_embedding = Embedding(tgt_token_to_id_map, tgt_emb_hparams)

        return src_embedding, tgt_embedding

    def process(self, raw_example: Tuple[List[str], List[str]]) -> \
            Tuple[List[str], List[str]]:
        # `_process` truncates sentences and appends BOS/EOS tokens.
        src_words = raw_example[0]
        if (self._src_max_seq_length is not None and
                len(src_words) > self._src_max_seq_length):
            if self._src_length_filter_mode is _LengthFilterMode.TRUNC:
                src_words = src_words[:self._src_max_seq_length]

        if self._src_bos_token is not None and self._src_bos_token != '':
            src_words.insert(0, self._src_bos_token)
        if self._src_eos_token is not None and self._src_eos_token != '':
            src_words.append(self._src_eos_token)

        # apply the transformations to source
        for transform in self._src_transforms:
            src_words = transform(src_words)

        tgt_words = raw_example[1]
        if (self._tgt_max_seq_length is not None and
                len(tgt_words) > self._tgt_max_seq_length):
            if self._tgt_length_filter_mode is _LengthFilterMode.TRUNC:
                tgt_words = tgt_words[:self._tgt_max_seq_length]

        if self._tgt_bos_token is not None and self._tgt_bos_token != '':
            tgt_words.insert(0, self._tgt_bos_token)
        if self._tgt_eos_token is not None and self._tgt_eos_token != '':
            tgt_words.append(self._tgt_eos_token)

        # apply the transformations to target
        for transform in self._tgt_transforms:
            tgt_words = transform(tgt_words)

        return src_words, tgt_words

    @staticmethod
    def _get_name_prefix(src_hparams, tgt_hparams):
        name_prefix = [
            src_hparams["data_name"], tgt_hparams["data_name"]]
        if name_prefix[0] == name_prefix[1]:
            raise ValueError("'data_name' of source and target "
                             "datasets cannot be the same.")
        return name_prefix

    def collate(self, examples: List[Tuple[List[str], List[str]]]) -> Batch:
        # For `PairedTextData`, each example is represented as a tuple of list
        # of strings.
        # `_collate` takes care of padding and numericalization.

        # If `pad_length` is `None`, pad to the longest sentence in the batch.
        src_examples = [example[0] for example in examples]
        source_ids = [self._src_vocab.map_tokens_to_ids_py(sent) for sent
                      in src_examples]
        source_ids, source_lengths = \
            padded_batch(source_ids,
                         self._src_pad_length,
                         pad_value=self._src_vocab.pad_token_id)
        src_pad_length = self._src_pad_length or max(source_lengths)
        src_examples = [
            sent + [''] * (src_pad_length - len(sent))
            if len(sent) < src_pad_length else sent
            for sent in src_examples
        ]

        source_ids = torch.from_numpy(source_ids)
        source_lengths = torch.tensor(source_lengths, dtype=torch.long)

        tgt_examples = [example[1] for example in examples]
        target_ids = [self._tgt_vocab.map_tokens_to_ids_py(sent) for sent
                      in tgt_examples]
        target_ids, target_lengths = \
            padded_batch(target_ids,
                         self._tgt_pad_length,
                         pad_value=self._tgt_vocab.pad_token_id)
        tgt_pad_length = self._tgt_pad_length or max(target_lengths)
        tgt_examples = [
            sent + [''] * (tgt_pad_length - len(sent))
            if len(sent) < tgt_pad_length else sent
            for sent in tgt_examples
        ]

        target_ids = torch.from_numpy(target_ids)
        target_lengths = torch.tensor(target_lengths, dtype=torch.long)

        return Batch(len(examples), source_text=src_examples,
                     source_text_ids=source_ids, source_length=source_lengths,
                     target_text=tgt_examples, target_text_ids=target_ids,
                     target_length=target_lengths)

    def list_items(self) -> List[str]:
        r"""Returns the list of item names that the data can produce.

        Returns:
            A list of strings.
        """
        items = ['text', 'text_ids', 'length']
        src_name = self._hparams.source_dataset['data_name']
        tgt_name = self._hparams.target_dataset['data_name']

        if src_name is not None:
            src_items = [src_name + '_' + item for item in items]
        else:
            src_items = items

        if tgt_name is not None:
            tgt_items = [tgt_name + '_' + item for item in items]
        else:
            tgt_items = items

        return src_items + tgt_items

    @property
    def vocab(self):
        r"""A pair instances of :class:`~texar.torch.data.Vocab` that are source
        and target vocabs, respectively.
        """
        return self._src_vocab, self._tgt_vocab

    @property
    def source_vocab(self):
        r"""The source vocab, an instance of :class:`~texar.torch.data.Vocab`.
        """
        return self._src_vocab

    @property
    def target_vocab(self):
        r"""The target vocab, an instance of :class:`~texar.torch.data.Vocab`.
        """
        return self._tgt_vocab

    @property
    def source_text_name(self):
        r"""The name for source text"""
        name = "{}_text".format(self.hparams.source_dataset["data_name"])
        return name

    @property
    def source_text_id_name(self):
        r"""The name for source text id"""
        name = "{}_text_ids".format(self.hparams.source_dataset["data_name"])
        return name

    @property
    def source_length_name(self):
        r"""The name for source length"""
        name = "{}_length".format(self.hparams.source_dataset["data_name"])
        return name

    @property
    def target_text_name(self):
        r"""The name for target text"""
        name = "{}_text".format(self.hparams.target_dataset["data_name"])
        return name

    @property
    def target_text_id_name(self):
        r"""The name for target text id"""
        name = "{}_text_ids".format(self.hparams.target_dataset["data_name"])
        return name

    @property
    def target_length_name(self):
        r"""The name for target length"""
        name = "{}_length".format(self.hparams.target_dataset["data_name"])
        return name

    def embedding_init_value(self):
        r"""A pair of `Tensor` containing the embedding values of source and
        target data loaded from file.
        """
        src_emb = self.hparams.source_dataset["embedding_init"]
        tgt_emb = self.hparams.target_dataser["embedding_init"]
        return src_emb, tgt_emb



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for rnn helpers.
"""

import unittest

import torch

from texar.torch.core.cell_wrappers import RNNCell, GRUCell, LSTMCell
from texar.torch.utils.rnn import (
    dynamic_rnn, reverse_sequence, bidirectional_dynamic_rnn)


class ReverseSequenceTest(unittest.TestCase):
    r"""Tests reverse_sequence.
    """

    def setUp(self):
        self.inputs = [[[10, 11], [12, 13], [14, 15], [16, 17], [18, 19]],
                       [[20, 21], [22, 23], [24, 25], [26, 27], [28, 29]],
                       [[30, 31], [32, 33], [34, 35], [36, 37], [38, 39]],
                       [[40, 41], [42, 43], [44, 45], [46, 47], [48, 49]]]
        self.inputs = torch.tensor(self.inputs)

    def test_reverse_sequence(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.reverse_sequence`.
        """
        seq_lengths_batch_first = torch.tensor([1, 2, 3, 4])
        output = reverse_sequence(inputs=self.inputs,
                                  seq_lengths=seq_lengths_batch_first,
                                  time_major=False)

        expect_out = [[[10, 11], [12, 13], [14, 15], [16, 17], [18, 19]],
                      [[22, 23], [20, 21], [24, 25], [26, 27], [28, 29]],
                      [[34, 35], [32, 33], [30, 31], [36, 37], [38, 39]],
                      [[46, 47], [44, 45], [42, 43], [40, 41], [48, 49]]]

        self.assertEqual(output.tolist(), expect_out)

        seq_lengths_time_first = torch.tensor([0, 1, 2, 3, 4])
        output = reverse_sequence(inputs=self.inputs,
                                  seq_lengths=seq_lengths_time_first,
                                  time_major=True)

        expect_out = [[[10, 11], [12, 13], [24, 25], [36, 37], [48, 49]],
                      [[20, 21], [22, 23], [14, 15], [26, 27], [38, 39]],
                      [[30, 31], [32, 33], [34, 35], [16, 17], [28, 29]],
                      [[40, 41], [42, 43], [44, 45], [46, 47], [18, 19]]]

        self.assertEqual(output.tolist(), expect_out)


class DynamicRNNTest(unittest.TestCase):
    r"""Tests dynamic_rnn.
    """

    def setUp(self):
        self._batch_size = 8
        self._max_time = 64
        self._input_size = 16
        self._hidden_size = 32
        self._rnn = RNNCell(self._input_size, self._hidden_size)
        self._lstm = LSTMCell(self._input_size, self._hidden_size)
        self._gru = GRUCell(self._input_size, self._hidden_size)

    def test_dynamic_rnn_basic(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.dynamic_rnn`.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)

        # RNN
        outputs, final_state = dynamic_rnn(self._rnn,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=None,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertEqual(final_state.shape, torch.Size([self._batch_size,
                                                        self._hidden_size]))
        # LSTM
        outputs, final_state = dynamic_rnn(self._lstm,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=None,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertIsInstance(final_state, tuple)
        self.assertEqual(final_state[0].shape, torch.Size([self._batch_size,
                                                           self._hidden_size]))
        self.assertEqual(final_state[1].shape, torch.Size([self._batch_size,
                                                           self._hidden_size]))

        # GRU
        outputs, final_state = dynamic_rnn(self._gru,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=None,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertEqual(final_state.shape, torch.Size([self._batch_size,
                                                        self._hidden_size]))

    def test_dynamic_rnn_time_major(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.dynamic_rnn`.
        """
        inputs = torch.rand(self._max_time, self._batch_size, self._input_size)

        # RNN
        outputs, final_state = dynamic_rnn(self._rnn,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=None,
                                           time_major=True)

        self.assertEqual(outputs.shape, torch.Size([self._max_time,
                                                    self._batch_size,
                                                    self._hidden_size]))
        self.assertEqual(final_state.shape, torch.Size([self._batch_size,
                                                        self._hidden_size]))
        # LSTM
        outputs, final_state = dynamic_rnn(self._lstm,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=None,
                                           time_major=True)

        self.assertEqual(outputs.shape, torch.Size([self._max_time,
                                                    self._batch_size,
                                                    self._hidden_size]))
        self.assertIsInstance(final_state, tuple)
        self.assertEqual(final_state[0].shape, torch.Size([self._batch_size,
                                                           self._hidden_size]))
        self.assertEqual(final_state[1].shape, torch.Size([self._batch_size,
                                                           self._hidden_size]))

        # GRU
        outputs, final_state = dynamic_rnn(self._gru,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=None,
                                           time_major=True)

        self.assertEqual(outputs.shape, torch.Size([self._max_time,
                                                    self._batch_size,
                                                    self._hidden_size]))
        self.assertEqual(final_state.shape, torch.Size([self._batch_size,
                                                        self._hidden_size]))

    def test_dynamic_rnn_sequence_length(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.dynamic_rnn`.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)
        sequence_length = [0, 43, 23, 63, 12, 54, 33, 8]

        # RNN
        outputs, final_state = dynamic_rnn(self._rnn,
                                           inputs,
                                           sequence_length=sequence_length,
                                           initial_state=None,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertEqual(final_state.shape, torch.Size([self._batch_size,
                                                        self._hidden_size]))

        # LSTM
        outputs, final_state = dynamic_rnn(self._lstm,
                                           inputs,
                                           sequence_length=sequence_length,
                                           initial_state=None,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertIsInstance(final_state, tuple)
        self.assertEqual(final_state[0].shape, torch.Size([self._batch_size,
                                                           self._hidden_size]))
        self.assertEqual(final_state[1].shape, torch.Size([self._batch_size,
                                                           self._hidden_size]))

        # GRU
        outputs, final_state = dynamic_rnn(self._gru,
                                           inputs,
                                           sequence_length=sequence_length,
                                           initial_state=None,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertEqual(final_state.shape, torch.Size([self._batch_size,
                                                        self._hidden_size]))

    def test_dynamic_rnn_initial_state(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.dynamic_rnn`.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)

        rnn_initial_state = torch.rand(self._batch_size, self._hidden_size)
        lstm_initial_state = (torch.rand(self._batch_size, self._hidden_size),
                              torch.rand(self._batch_size, self._hidden_size))

        # RNN
        outputs, final_state = dynamic_rnn(self._rnn,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=rnn_initial_state,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertEqual(final_state.shape, torch.Size([self._batch_size,
                                                        self._hidden_size]))

        # LSTM
        outputs, final_state = dynamic_rnn(self._lstm,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=lstm_initial_state,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertIsInstance(final_state, tuple)
        self.assertEqual(final_state[0].shape, torch.Size([self._batch_size,
                                                           self._hidden_size]))
        self.assertEqual(final_state[1].shape, torch.Size([self._batch_size,
                                                           self._hidden_size]))

        # GRU
        outputs, final_state = dynamic_rnn(self._gru,
                                           inputs,
                                           sequence_length=None,
                                           initial_state=rnn_initial_state,
                                           time_major=False)

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    self._hidden_size]))
        self.assertEqual(final_state.shape, torch.Size([self._batch_size,
                                                        self._hidden_size]))


class BidirectionalDynamicRNNTest(unittest.TestCase):
    r"""Tests bidirectional_dynamic_rnn.
    """

    def setUp(self):
        self._batch_size = 8
        self._max_time = 64
        self._input_size = 16
        self._hidden_size = 32
        self._rnn_fw = RNNCell(self._input_size, self._hidden_size)
        self._rnn_bw = RNNCell(self._input_size, self._hidden_size)
        self._lstm_fw = LSTMCell(self._input_size, self._hidden_size)
        self._lstm_bw = LSTMCell(self._input_size, self._hidden_size)
        self._gru_fw = GRUCell(self._input_size, self._hidden_size)
        self._gru_bw = GRUCell(self._input_size, self._hidden_size)

    def test_bidirectional_dynamic_rnn_basic(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.bidirectional_dynamic_rnn`.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)

        # RNN
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._rnn_fw,
            cell_bw=self._rnn_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertEqual(output_state[0].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))
        self.assertEqual(output_state[1].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))

        # LSTM
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._lstm_fw,
            cell_bw=self._lstm_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertIsInstance(output_state[0], tuple)
        self.assertEqual(output_state[0][0].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertEqual(output_state[0][1].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertIsInstance(output_state[1], tuple)
        self.assertEqual(output_state[1][0].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertEqual(output_state[1][1].shape,
                         torch.Size([self._batch_size, self._hidden_size]))

        # GRU
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._gru_fw,
            cell_bw=self._gru_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertEqual(output_state[0].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))
        self.assertEqual(output_state[1].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))

    def test_bidirectional_dynamic_rnn_time_major(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.bidirectional_dynamic_rnn`.
        """
        inputs = torch.rand(self._max_time, self._batch_size, self._input_size)

        # RNN
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._rnn_fw,
            cell_bw=self._rnn_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=True)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._max_time,
                                                       self._batch_size,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._max_time,
                                                       self._batch_size,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertEqual(output_state[0].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))
        self.assertEqual(output_state[1].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))

        # LSTM
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._lstm_fw,
            cell_bw=self._lstm_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=True)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._max_time,
                                                       self._batch_size,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._max_time,
                                                       self._batch_size,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertIsInstance(output_state[0], tuple)
        self.assertEqual(output_state[0][0].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertEqual(output_state[0][1].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertIsInstance(output_state[1], tuple)
        self.assertEqual(output_state[1][0].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertEqual(output_state[1][1].shape,
                         torch.Size([self._batch_size, self._hidden_size]))

        # GRU
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._gru_fw,
            cell_bw=self._gru_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=True)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._max_time,
                                                       self._batch_size,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._max_time,
                                                       self._batch_size,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertEqual(output_state[0].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))
        self.assertEqual(output_state[1].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))

    def test_bidirectional_dynamic_rnn_sequence_length(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.bidirectional_dynamic_rnn`.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)
        sequence_length = [0, 43, 23, 63, 12, 54, 33, 8]

        # RNN
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._rnn_fw,
            cell_bw=self._rnn_bw,
            inputs=inputs,
            sequence_length=sequence_length,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertEqual(output_state[0].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))
        self.assertEqual(output_state[1].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))

        # LSTM
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._lstm_fw,
            cell_bw=self._lstm_bw,
            inputs=inputs,
            sequence_length=sequence_length,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertIsInstance(output_state[0], tuple)
        self.assertEqual(output_state[0][0].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertEqual(output_state[0][1].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertIsInstance(output_state[1], tuple)
        self.assertEqual(output_state[1][0].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertEqual(output_state[1][1].shape,
                         torch.Size([self._batch_size, self._hidden_size]))

        # GRU
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._gru_fw,
            cell_bw=self._gru_bw,
            inputs=inputs,
            sequence_length=sequence_length,
            initial_state_fw=None,
            initial_state_bw=None,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertEqual(output_state[0].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))
        self.assertEqual(output_state[1].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))

    def test_bidirectional_dynamic_rnn_initial_state(self):
        r"""Tests :meth:`~texar.torch.utils.rnn.bidirectional_dynamic_rnn`.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)

        rnn_initial_state_fw = torch.rand(self._batch_size, self._hidden_size)
        rnn_initial_state_bw = torch.rand(self._batch_size, self._hidden_size)
        lstm_initial_state_fw = (torch.rand(self._batch_size,
                                            self._hidden_size),
                                 torch.rand(self._batch_size,
                                            self._hidden_size))
        lstm_initial_state_bw = (torch.rand(self._batch_size,
                                            self._hidden_size),
                                 torch.rand(self._batch_size,
                                            self._hidden_size))

        # RNN
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._rnn_fw,
            cell_bw=self._rnn_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=rnn_initial_state_fw,
            initial_state_bw=rnn_initial_state_bw,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertEqual(output_state[0].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))
        self.assertEqual(output_state[1].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))

        # LSTM
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._lstm_fw,
            cell_bw=self._lstm_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=lstm_initial_state_fw,
            initial_state_bw=lstm_initial_state_bw,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertIsInstance(output_state[0], tuple)
        self.assertEqual(output_state[0][0].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertEqual(output_state[0][1].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertIsInstance(output_state[1], tuple)
        self.assertEqual(output_state[1][0].shape,
                         torch.Size([self._batch_size, self._hidden_size]))
        self.assertEqual(output_state[1][1].shape,
                         torch.Size([self._batch_size, self._hidden_size]))

        # GRU
        outputs, output_state = bidirectional_dynamic_rnn(
            cell_fw=self._gru_fw,
            cell_bw=self._gru_bw,
            inputs=inputs,
            sequence_length=None,
            initial_state_fw=rnn_initial_state_fw,
            initial_state_bw=rnn_initial_state_bw,
            time_major=False)

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       self._hidden_size]))

        self.assertIsInstance(output_state, tuple)
        self.assertEqual(output_state[0].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))
        self.assertEqual(output_state[1].shape, torch.Size([self._batch_size,
                                                            self._hidden_size]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various data iterator classes.
"""

# pylint: disable=protected-access

from typing import (
    Dict, Iterable, Iterator, List, Optional, Sequence, Union, Mapping)

import pkg_resources
import torch
from torch import __version__ as _torch_version  # type: ignore
from torch.utils.data import DataLoader

from texar.torch.data.data.data_base import DatasetBase
from texar.torch.data.data.dataset_utils import Batch
from texar.torch.data.data.sampler import (
    SamplerBase, SequentialSampler, RandomSampler, BufferShuffleSampler,
    BatchingStrategy, DynamicBatchSampler)
from texar.torch.utils.types import MaybeSeq
from texar.torch.utils.utils import ceildiv, map_structure

_torch_version = pkg_resources.parse_version(_torch_version)

__all__ = [
    "DataIterator",
    "TrainTestDataIterator",
]

# `Dict` is invariant, `Mapping` is not.
DatasetsType = Union[Mapping[str, DatasetBase], MaybeSeq[DatasetBase]]


# pylint: disable=ungrouped-imports
if _torch_version >= pkg_resources.parse_version("1.2.0"):  # PyTorch 1.2.0 +
    from torch.utils.data._utils.pin_memory import (  # type: ignore
        pin_memory as _pin_memory)
elif _torch_version >= pkg_resources.parse_version("1.1.0"):  # PyTorch 1.1.0 +
    from torch.utils.data._utils.pin_memory import (  # type: ignore
        pin_memory_batch as _pin_memory)
else:
    from torch.utils.data.dataloader import (  # type: ignore
        pin_memory_batch as _pin_memory)


def move_memory(data, device):
    def _move_fn(x):
        if isinstance(x, torch.Tensor):
            return x.to(device=device, non_blocking=True)
        return x

    if isinstance(data, Batch):
        return Batch(len(data), batch={
            key: map_structure(_move_fn, value)
            for key, value in data.items()
        })
    return map_structure(_move_fn, data)


if _torch_version >= pkg_resources.parse_version("1.2.0"):  # PyTorch 1.2.0 +
    # PyTorch 1.2 split the `_DataLoaderIter` class into two:
    # `_SingleProcessDataLoaderIter` for when `num_workers == 0`, i.e. when
    # multi-processing is disabled; `_MultiProcessingDataLoaderIter` for
    # otherwise. The implementation is also slightly different from previous
    # releases.
    #
    # To keep compatibility, our iterator classes should be a subclass of both
    # PyTorch `_Single...`/`_Multi...` (for single/multi-process), and our own
    # `_Cache...`/`_Data...` (for caching/no caching). This results in four
    # different concrete classes, as this regex shows:
    # `_[SM]P(Cache)?DataLoaderIter`.
    #
    # We only expose `_DataLoaderIter` and `_CacheDataLoaderIter` to other
    # classes, and construct concrete classes in their `__new__` methods
    # depending on the value of `num_workers`. This is for compatibility with
    # previous versions, so we don't need to change other parts of the code.

    from texar.torch.data.data.data_iterators_utils import \
        TexarBaseDataLoaderIter as _BaseDataLoaderIter
    from texar.torch.data.data.data_iterators_utils import \
        TexarSingleProcessDataLoaderIter as _SingleProcessDataLoaderIter
    from texar.torch.data.data.data_iterators_utils import \
        TexarMultiProcessingDataLoaderIter as _MultiProcessingDataLoaderIter

    class _DataLoaderIter(_BaseDataLoaderIter):
        r"""Iterates once over the DataLoader's dataset. This is almost
        identical to PyTorch
        :class:`torch.utils.data.dataloader._BaseDataLoaderIter`, except that we
        check `allow_smaller_final_batch` here. This is because using
        `drop_last` in :class:`~torch.utils.data.sampler.BatchSampler` would
        cause the dataset to not load/process/cache certain elements from the
        final batch, which complicates the already complex logic.
        """

        def __new__(cls, loader: 'SingleDatasetIterator'):
            if loader.num_workers > 0:
                return super().__new__(_MPDataLoaderIter)
            else:
                return super().__new__(_SPDataLoaderIter)

        def __init__(self, loader: 'SingleDatasetIterator'):
            self.device = loader.device
            self._batch_size = loader.batch_size
            super().__init__(loader)

        def __next__(self):
            batch = super().__next__()
            # Drop smaller final batch according to settings. Note that
            # `_batch_size` could be None if dynamic batching is used.
            if (self._batch_size is not None and
                    batch.batch_size < self._batch_size and
                    not self.dataset.hparams.allow_smaller_final_batch):
                raise StopIteration
            if self.device is not None:
                batch = move_memory(batch, self.device)
            return batch

    class _SPDataLoaderIter(_DataLoaderIter, _SingleProcessDataLoaderIter):
        pass

    class _MPDataLoaderIter(_DataLoaderIter, _MultiProcessingDataLoaderIter):
        pass

    class _CacheDataLoaderIter(_BaseDataLoaderIter):
        r"""Iterates once over the DataLoader's dataset. This class is used when
        examples are processed and returned by worker processes. We need to
        record the corresponding indices of each batch, call
        :meth:`texar.torch.data.data.DatasetBase._add_cached_examples` to cache
        the processed examples, and return only the
        :class:`~texar.torch.data.data.Batch` instance to the user.
        """

        def __new__(cls, loader: 'SingleDatasetIterator'):
            if loader.num_workers > 0:
                return super().__new__(_MPCacheDataLoaderIter)
            else:
                return super().__new__(_SPCacheDataLoaderIter)

        def __init__(self, loader: 'SingleDatasetIterator'):
            self._indices_dict: Dict[int, List[int]] = {}
            self._batch_size = loader.batch_size
            self.device = loader.device
            super().__init__(loader)

    class _SPCacheDataLoaderIter(_CacheDataLoaderIter,
                                 _SingleProcessDataLoaderIter):
        def __next__(self):
            index = self._next_index()  # may raise StopIteration
            data = self.dataset_fetcher.fetch(index)  # may raise StopIteration
            if self.dataset._should_yield_raw_example:
                index = [idx[0] for idx in index]
            examples, data = data
            self.dataset._add_cached_examples(index, examples)
            if self.pin_memory:
                data = move_memory(_pin_memory(data), self.device)
            return data

    class _MPCacheDataLoaderIter(_CacheDataLoaderIter,
                                 _MultiProcessingDataLoaderIter):
        dataset: DatasetBase

        worker_queue_idx: int  # so that Pylint gives no errors

        def _try_put_index(self):
            assert self.tasks_outstanding < 2 * self.num_workers
            try:
                index = self._next_index()
            except StopIteration:
                return
            for _ in range(self.num_workers):  # find next active worker, if any
                worker_queue_idx = next(self.worker_queue_idx_cycle)
                if self.workers_status[worker_queue_idx]:
                    break
            else:
                # not found (i.e., didn't break)
                return

            self.index_queues[worker_queue_idx].put((self.send_idx, index))
            if self.dataset._should_yield_raw_example:
                index = [idx[0] for idx in index]
            self._indices_dict[self.send_idx] = index
            self.task_info[self.send_idx] = (worker_queue_idx,)
            self.tasks_outstanding += 1
            self.send_idx += 1

        def _process_data(self, batch):
            batch = super()._process_data(batch)
            indices = self._indices_dict[self.rcvd_idx - 1]
            del self._indices_dict[self.rcvd_idx - 1]
            examples, batch = batch
            self.dataset._add_cached_examples(indices, examples)
            return batch

        def __next__(self):
            batch = super().__next__()
            if (self._batch_size is not None and
                    batch.batch_size < self.dataset.batch_size and
                    not self.dataset.hparams.allow_smaller_final_batch):
                raise StopIteration
            batch = move_memory(batch, self.device)
            return batch
else:
    # PyTorch 1.1 and lower defines only the class `_DataLoaderIter` for
    # iterating over `DataLoader`.

    from torch.utils.data.dataloader import (  # type: ignore
        _DataLoaderIter as torch_DataLoaderIter)

    class _DataLoaderIter(torch_DataLoaderIter):  # type: ignore
        r"""Iterates once over the DataLoader's dataset. This is almost
        identical to PyTorch
        :class:`torch.utils.data.dataloader._DataLoaderIter`, except that we
        check `allow_smaller_final_batch` here. This is because using
        `drop_last` in :class:`~torch.utils.data.sampler.BatchSampler` would
        cause the dataset to not load/process/cache certain elements from the
        final batch, which complicates the already complex logic.
        """

        def __init__(self, loader: 'SingleDatasetIterator'):
            self._batch_size = loader.batch_size
            self.device = loader.device
            super().__init__(loader)

        def __next__(self):
            batch = super().__next__()
            # Drop smaller final batch according to settings. Note that
            # `_batch_size` could be None if dynamic batching is used.
            if (self._batch_size is not None and
                    batch.batch_size < self._batch_size and
                    not self.dataset.hparams.allow_smaller_final_batch):
                raise StopIteration
            batch = move_memory(batch, self.device)
            return batch

    class _CacheDataLoaderIter(torch_DataLoaderIter):  # type: ignore
        r"""Iterates once over the DataLoader's dataset. This class is used when
        examples are processed and returned by worker processes. We need to
        record the corresponding indices of each batch, call
        :meth:`texar.torch.data.data.DatasetBase._add_cached_examples` to cache
        the processed examples, and return only the
        :class:`~texar.torch.data.data.Batch` instance to the user.
        """
        dataset: DatasetBase

        worker_queue_idx: int  # so that Pylint gives no errors

        def __init__(self, loader: 'SingleDatasetIterator'):
            self._indices_dict: Dict[int, List[int]] = {}
            self._batch_size = loader.batch_size
            self.device = loader.device
            super().__init__(loader)

        def _put_indices(self):
            assert self.batches_outstanding < 2 * self.num_workers
            indices = next(self.sample_iter, None)
            if indices is None:
                return
            self.index_queues[self.worker_queue_idx].put(
                (self.send_idx, indices))
            if self.dataset._should_yield_raw_example:
                indices = [index[0] for index in indices]
            self._indices_dict[self.send_idx] = indices
            self.worker_queue_idx = ((self.worker_queue_idx + 1) %
                                     self.num_workers)
            self.batches_outstanding += 1
            self.send_idx += 1

        def _process_next_batch(self, batch):
            batch = super()._process_next_batch(batch)
            indices = self._indices_dict[self.rcvd_idx - 1]
            del self._indices_dict[self.rcvd_idx - 1]
            examples, batch = batch
            self.dataset._add_cached_examples(indices, examples)
            return batch

        def __next__(self):
            if self.num_workers == 0:  # same-process loading
                indices = next(self.sample_iter)  # may raise StopIteration
                batch = self.collate_fn([self.dataset[i] for i in indices])
                if self.dataset._should_yield_raw_example:
                    indices = [index[0] for index in indices]
                examples, batch = batch
                self.dataset._add_cached_examples(indices, examples)
                if self.pin_memory:
                    batch = _pin_memory(batch)
            else:
                batch = super().__next__()
            if (self._batch_size is not None and
                    batch.batch_size < self.dataset.batch_size and
                    not self.dataset.hparams.allow_smaller_final_batch):
                raise StopIteration
            batch = move_memory(batch, self.device)
            return batch


class SingleDatasetIterator(DataLoader):
    r"""Iterator for a single dataset. This iterator is based on the PyTorch
    :class:`~torch.utils.data.DataLoader` interface, with a custom shuffling
    routine. This class is used internally.

    Args:
        dataset: The dataset to iterator through. The dataset must be an
            instance of :class:`texar.torch.data.DatasetBase`, because
            configurations are read from the dataset `HParams`.
        batching_strategy: The batching strategy to use when performing dynamic
            batching. If `None`, fixed-sized batching is used.
        pin_memory: If `True`, tensors will be moved onto page-locked memory
            before returning. This argument is passed into the constructor for
            :torch_docs:`DataLoader <data.html#torch.utils.data.DataLoader>`.

            Defaults to `None`, which will set the value to `True` if the
            :class:`~texar.torch.data.DatasetBase` instance is set to use a CUDA
            device. Set to `True` or `False` to override this behavior.
    """
    dataset: DatasetBase

    def __init__(self, dataset: DatasetBase,
                 batching_strategy: Optional[BatchingStrategy] = None,
                 pin_memory: Optional[bool] = None):
        shuffle = dataset.hparams.shuffle
        shuffle_buffer_size = dataset.hparams.shuffle_buffer_size
        sampler: SamplerBase
        if shuffle and shuffle_buffer_size is not None:
            sampler = BufferShuffleSampler(dataset, shuffle_buffer_size)
        elif shuffle:
            sampler = RandomSampler(dataset)
        else:
            sampler = SequentialSampler(dataset)

        num_workers = dataset.hparams.num_parallel_calls
        collate_fn = dataset._collate_and_maybe_return

        is_cuda = dataset.device is not None and dataset.device.type == "cuda"
        if pin_memory is None:
            pin_memory = is_cuda
        self.device = None
        if pin_memory and is_cuda:
            self.device = dataset.device

        if batching_strategy is not None:
            batch_sampler = DynamicBatchSampler(
                dataset, sampler, batching_strategy)
            super().__init__(
                dataset, batch_sampler=batch_sampler,
                collate_fn=collate_fn, num_workers=num_workers,
                pin_memory=pin_memory)
        else:
            super().__init__(
                dataset, batch_size=dataset.batch_size, drop_last=False,
                sampler=sampler, collate_fn=collate_fn, num_workers=num_workers,
                pin_memory=pin_memory)

    def __iter__(self):
        if self.dataset._should_return_processed_examples:
            # Accepts processed examples from workers and add to dataset cache.
            return _CacheDataLoaderIter(self)
        else:
            return _DataLoaderIter(self)

    def __len__(self):
        if self.batch_size is None:
            raise TypeError("__len__ not supported for dynamic batching")
        data_length = len(self.dataset)  # may throw TypeError
        if self.dataset.hparams.allow_smaller_final_batch:
            return ceildiv(data_length, self.batch_size)
        return data_length // self.batch_size


class DataIterator:
    r"""Data iterator that switches and iterates through multiple datasets.

    This is a wrapper of :class:`~texar.torch.data.SingleDatasetIterator`.

    Args:
        datasets: Datasets to iterate through. This can be:

            - A single instance of :class:`~texar.torch.data.DatasetBase`.
            - A `dict` that maps dataset name to instances of
              :class:`~texar.torch.data.DatasetBase`.
            - A `list` of instances of :class:`texar.torch.data.DatasetBase`.
              The name of instances (:attr:`texar.torch.data.DatasetBase.name`)
              must be unique.

        batching_strategy: The batching strategy to use when performing dynamic
            batching. If `None`, fixed-sized batching is used.
        pin_memory: If `True`, tensors will be moved onto page-locked memory
            before returning. This argument is passed into the constructor for
            :torch_docs:`DataLoader <data.html#torch.utils.data.DataLoader>`.

            Defaults to `None`, which will set the value to `True` if the
            :class:`~texar.torch.data.DatasetBase` instance is set to use a CUDA
            device. Set to `True` or `False` to override this behavior.

    Example:

        Create an iterator over two datasets and generating fixed-sized batches:

        .. code-block:: python

            train_data = MonoTextData(hparams_train)
            test_data = MonoTextData(hparams_test)
            iterator = DataIterator({'train': train_data, 'test': test_data})

            for epoch in range(200): # Run 200 epochs of train/test
                # Starts iterating through training data from the beginning.
                iterator.switch_to_dataset('train')
                for batch in iterator:
                    ... # Do training with the batch.

                # Starts iterating through test data from the beginning
                for batch in iterator.get_iterator('test'):
                    ... # Do testing with the batch.

        Dynamic batching based on total number of tokens:

        .. code-block:: python

            iterator = DataIterator(
                {'train': train_data, 'test': test_data},
                batching_strategy=TokenCountBatchingStrategy(max_tokens=1000))

        Dynamic batching with custom strategy (e.g. total number of tokens in
        examples from :class:`~texar.torch.data.PairedTextData`, including
        padding):

        .. code-block:: python

            class CustomBatchingStrategy(BatchingStrategy):
                def __init__(self, max_tokens: int):
                    self.max_tokens = max_tokens
                    self.reset_batch()

                def reset_batch(self) -> None:
                    self.max_src_len = 0
                    self.max_tgt_len = 0
                    self.cur_batch_size = 0

                def add_example(self, ex: Tuple[List[str], List[str]]) -> bool:
                    max_src_len = max(self.max_src_len, len(ex[0]))
                    max_tgt_len = max(self.max_tgt_len, len(ex[0]))
                    if (max(max_src_len + max_tgt_len) *
                            (self.cur_batch_size + 1) > self.max_tokens):
                        return False
                    self.max_src_len = max_src_len
                    self.max_tgt_len = max_tgt_len
                    self.cur_batch_size += 1
                    return True

            iterator = DataIterator(
                {'train': train_data, 'test': test_data},
                batching_strategy=CustomBatchingStrategy(max_tokens=1000))
    """

    # TODO: Think about whether we should support save/load.

    def __init__(self, datasets: DatasetsType,
                 batching_strategy: Optional[BatchingStrategy] = None,
                 pin_memory: Optional[bool] = None):
        self._default_dataset_name = 'data'
        if isinstance(datasets, DatasetBase):
            datasets = {self._default_dataset_name: datasets}
        elif isinstance(datasets, Sequence):
            if any(not isinstance(d, DatasetBase) for d in datasets):
                raise ValueError("`datasets` must be an non-empty list of "
                                 "`texar.torch.data.DatasetBase` instances.")
            num_datasets = len(datasets)
            datasets = {d.name: d for d in datasets}
            if len(datasets) < num_datasets:
                raise ValueError("Names of datasets must be unique.")

        _datasets = {
            name: SingleDatasetIterator(dataset, batching_strategy, pin_memory)
            for name, dataset in datasets.items()}
        self._datasets = _datasets

        if len(self._datasets) <= 0:
            raise ValueError("`datasets` must not be empty.")

        self._current_dataset_name: Optional[str] = None

    @property
    def num_datasets(self) -> int:
        r"""Number of datasets.
        """
        return len(self._datasets)

    @property
    def dataset_names(self) -> List[str]:
        r"""A list of dataset names.
        """
        return list(self._datasets.keys())

    def _validate_dataset_name(self, dataset_name: Optional[str]) -> str:
        r"""Validate the provided dataset name, and return the validated name.
        """
        if dataset_name is None:
            if self.num_datasets > 1:
                raise ValueError("`dataset_name` is required if there are "
                                 "more than one datasets.")
            dataset_name = next(iter(self._datasets))
        if dataset_name not in self._datasets:
            raise ValueError("Dataset not found: ", dataset_name)
        return dataset_name

    def switch_to_dataset(self, dataset_name: Optional[str] = None):
        r"""Re-initializes the iterator of a given dataset and starts iterating
        over the dataset (from the beginning).

        Args:
            dataset_name (optional): Name of the dataset. If not provided,
                there must be only one Dataset.
        """
        self._current_dataset_name = self._validate_dataset_name(dataset_name)

    def get_iterator(self,
                     dataset_name: Optional[str] = None) -> Iterator[Batch]:
        r"""Re-initializes the iterator of a given dataset and starts iterating
        over the dataset (from the beginning).

        Args:
            dataset_name (optional): Name of the dataset. If not provided,
                there must be only one Dataset.
        """
        if dataset_name is not None or self._current_dataset_name is None:
            dataset_name = self._validate_dataset_name(dataset_name)
        elif self._current_dataset_name is not None:
            dataset_name = self._current_dataset_name
        else:
            raise ValueError("No dataset is selected.")

        return iter(self._datasets[dataset_name])

    def __iter__(self) -> Iterator[Batch]:
        r"""Returns the iterator for the currently selected or default dataset.
        """
        return self.get_iterator()

    def __len__(self):
        return len(self._datasets[self._validate_dataset_name(None)])


class TrainTestDataIterator(DataIterator):
    r"""Data iterator that alternates between training, validation, and test
    datasets.

    :attr:`train`, :attr:`val`, and :attr:`test` are instances of
    :class:`~texar.torch.data.DatasetBase`. At least one of them must be
    provided.

    This is a wrapper of :class:`~texar.torch.data.DataIterator`.

    Args:
        train (optional): Training data.
        val (optional): Validation data.
        test (optional): Test data.
        batching_strategy: The batching strategy to use when performing dynamic
            batching. If `None`, fixed-sized batching is used.
        pin_memory: If `True`, tensors will be moved onto page-locked memory
            before returning. This argument is passed into the constructor for
            :torch_docs:`DataLoader <data.html#torch.utils.data.DataLoader>`.

            Defaults to `None`, which will set the value to `True` if the
            :class:`~texar.torch.data.DatasetBase` instance is set to use a CUDA
            device. Set to `True` or `False` to override this behavior.

    Example:

        .. code-block:: python

            train_data = MonoTextData(hparams_train)
            val_data = MonoTextData(hparams_val)
            iterator = TrainTestDataIterator(train=train_data, val=val_data)

            for epoch in range(200): # Run 200 epochs of train/val
                # Starts iterating through training data from the beginning.
                iterator.switch_to_train_data(sess)
                for batch in iterator:
                    ... # Do training with the batch.

                # Starts iterating through val data from the beginning.
                for batch in iterator.get_val_iterator():
                    ... # Do validation on the batch.
    """

    def __init__(self, train: Optional[DatasetBase] = None,
                 val: Optional[DatasetBase] = None,
                 test: Optional[DatasetBase] = None,
                 batching_strategy: Optional[BatchingStrategy] = None,
                 pin_memory: Optional[bool] = None):
        dataset_dict = {}
        self._train_name = 'train'
        self._val_name = 'val'
        self._test_name = 'test'
        if train is not None:
            dataset_dict[self._train_name] = train
        if val is not None:
            dataset_dict[self._val_name] = val
        if test is not None:
            dataset_dict[self._test_name] = test
        if len(dataset_dict) == 0:
            raise ValueError("At least one of `train`, `val`, and `test` "
                             "must be provided.")

        super().__init__(dataset_dict, batching_strategy, pin_memory)

    def switch_to_train_data(self) -> None:
        r"""Switch to training data."""
        if self._train_name not in self._datasets:
            raise ValueError("Training data not provided.")
        self.switch_to_dataset(self._train_name)

    def switch_to_val_data(self) -> None:
        r"""Switch to validation data."""
        if self._val_name not in self._datasets:
            raise ValueError("Validation data not provided.")
        self.switch_to_dataset(self._val_name)

    def switch_to_test_data(self) -> None:
        r"""Switch to test data."""
        if self._test_name not in self._datasets:
            raise ValueError("Test data not provided.")
        self.switch_to_dataset(self._test_name)

    def get_train_iterator(self) -> Iterable[Batch]:
        r"""Obtain an iterator over training data."""
        if self._train_name not in self._datasets:
            raise ValueError("Training data not provided.")
        return self.get_iterator(self._train_name)

    def get_val_iterator(self) -> Iterable[Batch]:
        r"""Obtain an iterator over validation data."""
        if self._val_name not in self._datasets:
            raise ValueError("Validation data not provided.")
        return self.get_iterator(self._val_name)

    def get_test_iterator(self) -> Iterable[Batch]:
        r"""Obtain an iterator over test data."""
        if self._test_name not in self._datasets:
            raise ValueError("Test data not provided.")
        return self.get_iterator(self._test_name)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions related to data types.
"""

from typing import Any, Dict, Optional, Union

import numpy as np
import torch

from texar.torch.hyperparams import HParams

__all__ = [
    'torch_bool',
    'get_numpy_dtype',
    'is_str',
    'is_callable',
    'get_supported_scalar_types',
    'maybe_hparams_to_dict',
    'compat_as_text',
]

# `torch.bool` exists in PyTorch 1.1, but the default type for comparisons
# is still `torch.uint8`.
torch_bool = (torch.empty(()) < 0).dtype

DTYPE_MAP = {
    np.float32: ['float32', 'float', 'tf.float32', 'torch.float',
                 'torch.float32', float, np.float32, torch.float32],
    np.float64: ['float64', 'tf.float64', 'torch.float64', np.float64,
                 np.float_, torch.float64],
    np.float16: ['float16', 'tf.float16', 'torch.float16', np.float16,
                 torch.float16],
    np.int32: ['int', 'int32', 'tf.int32', 'torch.int', 'torch.int32', int,
               np.int32, torch.int32],
    np.int64: ['int64', 'tf.int64', 'torch.int64', np.int64, np.int_,
               torch.int64],
    np.int16: ['int16', 'tf.int16', 'torch.int16', np.int16, torch.int16],
    np.int8: ['int8', 'char', 'tf.int8', 'torch.int8', np.int8, torch.int8],
    np.uint8: ['uint8', 'tf.uint8', 'torch.uint8', np.uint8, torch.uint8],
    np.bool_: ['bool', 'tf.bool', 'torch.bool', bool, np.bool, np.bool_,
               torch_bool],
    np.str_: ['string', 'str', 'tf.string', str, np.str, np.str_],
    np.bytes_: ['bytes', 'np.bytes', bytes, np.bytes_]
}


def get_numpy_dtype(dtype: Union[str, type]):
    r"""Returns equivalent NumPy dtype.

    Args:
        dtype: A str, Python numeric or string type, NumPy data type, or
            PyTorch dtype.

    Returns:
        The corresponding NumPy dtype.
    """
    for np_dtype, valid_values in DTYPE_MAP.items():
        if dtype in valid_values:
            return np_dtype

    raise ValueError(
        f"Unsupported conversion from type {dtype!s} to NumPy dtype")


def is_callable(x):
    r"""Return `True` if :attr:`x` is callable.
    """
    return callable(x)


def is_str(x):
    r"""Returns `True` if :attr:`x` is either a str or unicode.
    Returns `False` otherwise.
    """
    return isinstance(x, str)


def get_supported_scalar_types():
    r"""Returns a list of scalar types supported.
    """
    types = []
    for key, value in DTYPE_MAP.items():
        if key not in {np.str_, np.bytes_}:
            types.extend(value)

    return types


def maybe_hparams_to_dict(hparams: Optional[Union[HParams, Dict[str, Any]]]) \
        -> Optional[Dict[str, Any]]:
    r"""If :attr:`hparams` is an instance of :class:`~texar.torch.HParams`,
    converts it to a ``dict`` and returns. If :attr:`hparams` is a ``dict``,
    returns as is.

    Args:
        hparams: The :class:`~texar.torch.HParams` instance to convert.

    Returns:
        dict: The corresponding ``dict`` instance
    """
    if hparams is None:
        return None
    if isinstance(hparams, dict):
        return hparams
    return hparams.todict()


def _maybe_list_to_array(str_list, dtype_as):
    if isinstance(dtype_as, (list, tuple)):
        return type(dtype_as)(str_list)
    elif isinstance(dtype_as, np.ndarray):
        return np.array(str_list)
    else:
        return str_list


def _as_text(bytes_or_text, encoding='utf-8'):
    r"""Returns the given argument as a unicode string.

    Adapted from ``tensorflow.compat.as_text``.

    Args:
        bytes_or_text: A ``bytes``, ``str``, or ``unicode`` object.
            encoding: A string indicating the charset for decoding unicode.

    Returns:
        A ``unicode`` (Python 2) or ``str`` (Python 3) object.

    Raises:
        TypeError: If ``bytes_or_text`` is not a binary or unicode string.
    """
    if isinstance(bytes_or_text, str):
        return bytes_or_text
    elif isinstance(bytes_or_text, bytes):
        return bytes_or_text.decode(encoding)
    else:
        raise TypeError(
            f"Expected binary or unicode string, got {bytes_or_text!r}")


def compat_as_text(str_):
    r"""Converts strings into ``unicode`` (Python 2) or ``str`` (Python 3).

    Args:
        str\_: A string or other data types convertible to string, or an
            `n`-D numpy array or (possibly nested) list of such elements.

    Returns:
        The converted strings of the same structure/shape as :attr:`str_`.
    """

    def _recur_convert(s):
        if isinstance(s, (list, tuple, np.ndarray)):
            s_ = [_recur_convert(si) for si in s]
            return _maybe_list_to_array(s_, s)
        else:
            try:
                return _as_text(s)
            except TypeError:
                return _as_text(str(s))

    text = _recur_convert(str_)

    return text



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
RoBERTa encoder.
"""

from typing import Optional, Union

import torch

from texar.torch.modules.encoders.bert_encoder import BERTEncoder
from texar.torch.modules.pretrained.roberta import \
    PretrainedRoBERTaMixin

__all__ = [
    "RoBERTaEncoder",
]


class RoBERTaEncoder(PretrainedRoBERTaMixin, BERTEncoder):
    r"""RoBERTa Transformer for encoding sequences. Please see
    :class:`~texar.torch.modules.PretrainedRoBERTaMixin` for a brief description
    of RoBERTa.

    This module basically stacks
    :class:`~texar.torch.modules.WordEmbedder`,
    :class:`~texar.torch.modules.PositionEmbedder`,
    :class:`~texar.torch.modules.TransformerEncoder` and a dense
    pooler.

    Args:
        pretrained_model_name (optional): a `str`, the name
            of pre-trained model (e.g., ``roberta-base``). Please refer to
            :class:`~texar.torch.modules.PretrainedRoBERTaMixin` for
            all supported models.
            If `None`, the model name in :attr:`hparams` is used.
        cache_dir (optional): the path to a folder in which the
            pre-trained models will be cached. If `None` (default),
            a default directory (``texar_data`` folder under user's home
            directory) will be used.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure
            and default values.
    """

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        * The encoder arch is determined by the constructor argument
          :attr:`pretrained_model_name` if it's specified. In this case,
          `hparams` are ignored.
        * Otherwise, the encoder arch is determined by
          `hparams['pretrained_model_name']` if it's specified. All other
          configurations in `hparams` are ignored.
        * If the above two are `None`, the encoder arch is defined by the
          configurations in `hparams` and weights are randomly initialized.

        .. code-block:: python

            {
                "pretrained_model_name": "roberta-base",
                "embed": {
                    "dim": 768,
                    "name": "word_embeddings"
                },
                "vocab_size": 50265,
                "position_embed": {
                    "dim": 768,
                    "name": "position_embeddings"
                },
                "position_size": 514,

                "encoder": {
                    "dim": 768,
                    "embedding_dropout": 0.1,
                    "multihead_attention": {
                        "dropout_rate": 0.1,
                        "name": "self",
                        "num_heads": 12,
                        "num_units": 768,
                        "output_dim": 768,
                        "use_bias": True
                    },
                    "name": "encoder",
                    "num_blocks": 12,
                    "eps": 1e-12,
                    "poswise_feedforward": {
                        "layers": [
                            {
                                "kwargs": {
                                    "in_features": 768,
                                    "out_features": 3072,
                                    "bias": True
                                },
                                "type": "Linear"
                            },
                            {"type": "BertGELU"},
                            {
                                "kwargs": {
                                    "in_features": 3072,
                                    "out_features": 768,
                                    "bias": True
                                },
                                "type": "Linear"
                            }
                        ]
                    },
                    "residual_dropout": 0.1,
                    "use_bert_config": True
                    },
                "hidden_size": 768,
                "initializer": None,
                "name": "roberta_encoder",
            }

        Here:

        The default parameters are values for RoBERTa-Base model.

        `"pretrained_model_name"`: str or None
            The name of the pre-trained RoBERTa model. If None, the model
            will be randomly initialized.

        `"embed"`: dict
            Hyperparameters for word embedding layer.

        `"vocab_size"`: int
            The vocabulary size of `inputs` in RoBERTa model.

        `"position_embed"`: dict
            Hyperparameters for position embedding layer.

        `"position_size"`: int
            The maximum sequence length that this model might ever be used with.

        `"encoder"`: dict
            Hyperparameters for the TransformerEncoder.
            See :func:`~texar.torch.modules.TransformerEncoder.default_hparams`
            for details.

        `"hidden_size"`: int
            Size of the pooler dense layer.

        `"eps"`: float
            Epsilon values for layer norm layers.

        `"initializer"`: dict, optional
            Hyperparameters of the default initializer that initializes
            variables created in this module.
            See :func:`~texar.torch.core.get_initializer` for details.

        `"name"`: str
            Name of the module.
        """

        return {
            'pretrained_model_name': 'roberta-base',
            'embed': {
                'dim': 768,
                'name': 'word_embeddings'
            },
            'vocab_size': 50265,
            'position_embed': {
                'dim': 768,
                'name': 'position_embeddings'
            },
            'position_size': 514,

            'encoder': {
                'dim': 768,
                'embedding_dropout': 0.1,
                'multihead_attention': {
                    'dropout_rate': 0.1,
                    'name': 'self',
                    'num_heads': 12,
                    'num_units': 768,
                    'output_dim': 768,
                    'use_bias': True
                },
                'name': 'encoder',
                'num_blocks': 12,
                'eps': 1e-12,
                'poswise_feedforward': {
                    'layers': [
                        {
                            'kwargs': {
                                'in_features': 768,
                                'out_features': 3072,
                                'bias': True
                            },
                            'type': 'Linear'
                        },
                        {"type": "BertGELU"},
                        {
                            'kwargs': {
                                'in_features': 3072,
                                'out_features': 768,
                                'bias': True
                            },
                            'type': 'Linear'
                        }
                    ]
                },
                'residual_dropout': 0.1,
                'use_bert_config': True
            },
            'hidden_size': 768,
            'initializer': None,
            'name': 'roberta_encoder',
            '@no_typecheck': ['pretrained_model_name']
        }

    def forward(self,  # type: ignore
                inputs: Union[torch.Tensor, torch.LongTensor],
                sequence_length: Optional[torch.LongTensor] = None,
                segment_ids: Optional[torch.LongTensor] = None):
        r"""Encodes the inputs. Differing from the standard BERT, the RoBERTa
        model does not use segmentation embedding. As a result, RoBERTa does not
        require `segment_ids` as an input.

        Args:
            inputs: Either a **2D Tensor** of shape `[batch_size, max_time]`,
                containing the ids of tokens in input sequences, or
                a **3D Tensor** of shape `[batch_size, max_time, vocab_size]`,
                containing soft token ids (i.e., weights or probabilities)
                used to mix the embedding vectors.
            sequence_length (optional): A 1D Tensor of shape `[batch_size]`.
                Input tokens beyond respective sequence lengths are masked
                out automatically.

        Returns:
            A pair :attr:`(outputs, pooled_output)`

            - :attr:`outputs`:  A Tensor of shape
              `[batch_size, max_time, dim]` containing the encoded vectors.

            - :attr:`pooled_output`: A Tensor of size
              `[batch_size, hidden_size]` which is the output of a pooler
              pre-trained on top of the hidden state associated to the first
              character of the input (`CLS`), see RoBERTa's paper.
        """
        if segment_ids is not None:
            raise ValueError("segment_ids should be None in RoBERTaEncoder.")

        output, pooled_output = super().forward(inputs=inputs,
                                                sequence_length=sequence_length,
                                                segment_ids=None)

        return output, pooled_output



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various RNN decoders.
"""

from typing import Callable, Dict, NamedTuple, Optional, Tuple, Union

import torch
from torch import nn

from texar.torch.core import layers
from texar.torch.core.attention_mechanism import (
    AttentionMechanism, AttentionWrapperState)
from texar.torch.core.cell_wrappers import (
    AttentionWrapper, HiddenState, RNNCellBase)
from texar.torch.modules.decoders import decoder_helpers
from texar.torch.modules.decoders.decoder_base import (
    TokenEmbedder, TokenPosEmbedder)
from texar.torch.modules.decoders.decoder_helpers import Helper
from texar.torch.modules.decoders.rnn_decoder_base import RNNDecoderBase
from texar.torch.utils import utils
from texar.torch.utils.beam_search import beam_search
from texar.torch.utils.types import MaybeList, MaybeTuple
from texar.torch.utils.utils import check_or_get_instance, get_function

__all__ = [
    'BasicRNNDecoderOutput',
    'AttentionRNNDecoderOutput',
    'BasicRNNDecoder',
    'AttentionRNNDecoder',
]


class BasicRNNDecoderOutput(NamedTuple):
    r"""The outputs of :class:`~texar.torch.modules.BasicRNNDecoder` that
    include both RNN outputs and sampled IDs at each step. This is also used to
    store results of all the steps after decoding the whole sequence.
    """
    logits: torch.Tensor
    r"""The outputs of RNN (at each step/of all steps) by applying the
    output layer on cell outputs. For example, in
    :class:`~texar.torch.modules.BasicRNNDecoder` with default hyperparameters,
    this is a :tensor:`Tensor` of shape ``[batch_size, max_time, vocab_size]``
    after decoding the whole sequence."""
    sample_id: torch.LongTensor
    r"""The sampled results (at each step/of all steps). For example, in
    :class:`~texar.torch.modules.BasicRNNDecoder` with decoding strategy of
    ``"train_greedy"``, this is a :tensor:`LongTensor` of shape
    ``[batch_size, max_time]`` containing the sampled token indices of all
    steps. Note that the shape of ``sample_id`` is different for different
    decoding strategy or helper. Please refer to
    :class:`~texar.torch.modules.Helper` for the detailed information."""
    cell_output: torch.Tensor
    r"""The output of RNN cell (at each step/of all steps). This contains the
    results prior to the output layer. For example, in
    :class:`~texar.torch.modules.BasicRNNDecoder` with default hyperparameters,
    this is a :tensor:`Tensor` of shape
    ``[batch_size, max_time, cell_output_size]`` after decoding the whole
    sequence."""


class AttentionRNNDecoderOutput(NamedTuple):
    r"""The outputs of :class:`~texar.torch.modules.AttentionRNNDecoder` that
    additionally includes attention results.
    """
    logits: torch.Tensor
    r"""The outputs of RNN (at each step/of all steps) by applying the
    output layer on cell outputs. For example, in
    :class:`~texar.torch.modules.AttentionRNNDecoder` with default
    hyperparameters, this is a :tensor:`Tensor` of shape
    ``[batch_size, max_time, vocab_size]`` after decoding the whole sequence."""
    sample_id: torch.LongTensor
    r"""The sampled results (at each step/of all steps). For example, in
    :class:`~texar.torch.modules.AttentionRNNDecoder` with decoding strategy of
    ``"train_greedy"``, this is a :tensor:`LongTensor` of shape
    ``[batch_size, max_time]`` containing the sampled token indices of all
    steps. Note that the shape of ``sample_id`` is different for different
    decoding strategy or helper. Please refer to
    :class:`~texar.torch.modules.Helper` for the detailed information."""
    cell_output: torch.Tensor
    r"""The output of RNN cell (at each step/of all steps). This contains the
    results prior to the output layer. For example, in
    :class:`~texar.torch.modules.AttentionRNNDecoder` with default
    hyperparameters, this is a :tensor:`Tensor` of shape
    ``[batch_size, max_time, cell_output_size]`` after decoding the whole
    sequence."""
    attention_scores: MaybeTuple[torch.Tensor]
    r"""A single or tuple of `Tensor(s)` containing the alignments emitted (at
    the previous time step/of all time steps) for each attention mechanism."""
    attention_context: torch.Tensor
    r"""The attention emitted (at the previous time step/of all time steps)."""


class BasicRNNDecoder(RNNDecoderBase[HiddenState, BasicRNNDecoderOutput]):
    r"""Basic RNN decoder.

    Args:
        input_size (int): Dimension of input embeddings.
        vocab_size (int, optional): Vocabulary size. Required if
            :attr:`output_layer` is `None`.
        token_embedder: An instance of :torch_nn:`Module`, or a function taking
            a :tensor:`LongTensor` ``tokens`` as argument. This is the embedder
            called in :meth:`embed_tokens` to convert input tokens to
            embeddings.
        token_pos_embedder: An instance of :torch_nn:`Module`, or a function
            taking two :tensor:`LongTensor`\ s ``tokens`` and ``positions`` as
            argument. This is the embedder called in :meth:`embed_tokens` to
            convert input tokens with positions to embeddings.

            .. note::
                Only one among :attr:`token_embedder` and
                :attr:`token_pos_embedder` should be specified. If neither is
                specified, you must subclass :class:`BasicRNNDecoder` and
                override :meth:`embed_tokens`.
        cell (RNNCellBase, optional): An instance of
            :class:`~texar.torch.core.cell_wrappers.RNNCellBase`. If `None`
            (default), a cell is created as specified in :attr:`hparams`.
        output_layer (optional): An instance of :torch_nn:`Module`. Apply to
            the RNN cell output to get logits. If `None`, a :torch_nn:`Linear`
            layer is used with output dimension set to :attr:`vocab_size`.
            Set ``output_layer`` to :func:`~texar.torch.core.identity` if you do
            not want to have an output layer after the RNN cell outputs.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    See :meth:`~texar.torch.modules.RNNDecoderBase.forward` for the inputs and
    outputs of the decoder. The decoder returns
    ``(outputs, final_state, sequence_lengths)``, where ``outputs`` is an
    instance of :class:`~texar.torch.modules.BasicRNNDecoderOutput`.

    Example:
        .. code-block:: python

            embedder = WordEmbedder(vocab_size=data.vocab.size)
            decoder = BasicRNNDecoder(vocab_size=data.vocab.size)
            # Training loss
            outputs, _, _ = decoder(
                decoding_strategy='train_greedy',
                inputs=embedder(data_batch['text_ids']),
                sequence_length=data_batch['length']-1)
            loss = tx.losses.sequence_sparse_softmax_cross_entropy(
                labels=data_batch['text_ids'][:, 1:],
                logits=outputs.logits,
                sequence_length=data_batch['length']-1)

            # Create helper
            helper = decoder.create_helper(
                decoding_strategy='infer_sample',
                start_tokens=[data.vocab.bos_token_id]*100,
                end_token=data.vocab.eos.token_id,
                embedding=embedder)

            # Inference sample
            outputs, _, _ = decoder(
                helper=helerp,
                max_decoding_length=60)

            sample_text = tx.utils.map_ids_to_strs(
                outputs.sample_id, data.vocab)
            print(sample_text)
            # [
            #   the first sequence sample .
            #   the second sequence sample .
            #   ...
            # ]
    """

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "rnn_cell": default_rnn_cell_hparams(),
                "max_decoding_length_train": None,
                "max_decoding_length_infer": None,
                "helper_train": {
                    "type": "TrainingHelper",
                    "kwargs": {}
                }
                "helper_infer": {
                    "type": "SampleEmbeddingHelper",
                    "kwargs": {}
                }
                "name": "basic_rnn_decoder"
            }

        Here:

        `"rnn_cell"`: dict
            A dictionary of RNN cell hyperparameters. Ignored if
            :attr:`cell` is given to the decoder constructor.
            The default value is defined in
            :func:`~texar.torch.core.default_rnn_cell_hparams`.

        `"max_decoding_length_train"`: int or None
            Maximum allowed number of decoding steps in training mode. If
            `None` (default), decoding is performed until fully done, e.g.,
            encountering the ``<EOS>`` token. Ignored if
            ``"max_decoding_length"`` is not `None` given when calling the
            decoder.

        `"max_decoding_length_infer"`: int or None
            Same as ``"max_decoding_length_train"`` but for inference mode.

        `"helper_train"`: dict
            The hyperparameters of the helper used in training.
            ``"type"`` can be a helper class, its name or module path, or a
            helper instance. If a class name is given, the class must be
            from module :mod:`texar.torch.modules`, or
            :mod:`texar.torch.custom`. This is used only when both
            ``"decoding_strategy"`` and ``"helper"`` arguments are `None` when
            calling the decoder. See
            :meth:`~texar.torch.modules.RNNDecoderBase.forward` for more
            details.

        `"helper_infer"`: dict
            Same as ``"helper_train"`` but during inference mode.

        `"name"`: str
            Name of the decoder.
            The default value is ``"basic_rnn_decoder"``.
        """
        hparams = RNNDecoderBase.default_hparams()
        hparams['name'] = 'basic_rnn_decoder'
        return hparams

    def step(self, helper: Helper, time: int, inputs: torch.Tensor,
             state: Optional[HiddenState]) \
            -> Tuple[BasicRNNDecoderOutput, HiddenState]:
        cell_outputs, cell_state = self._cell(inputs, state)
        logits = self._output_layer(cell_outputs)
        sample_ids = helper.sample(time=time, outputs=logits)
        next_state = cell_state
        outputs = BasicRNNDecoderOutput(logits, sample_ids, cell_outputs)
        return outputs, next_state

    def next_inputs(self, helper: Helper, time: int,
                    outputs: BasicRNNDecoderOutput) -> \
            Tuple[torch.Tensor, torch.ByteTensor]:
        finished, next_inputs = helper.next_inputs(
            self.embed_tokens, time, outputs.logits, outputs.sample_id)
        return next_inputs, finished

    @property
    def output_size(self):
        r"""Output size of one step.
        """
        return self._cell.hidden_size


class AttentionRNNDecoder(RNNDecoderBase[AttentionWrapperState,
                                         AttentionRNNDecoderOutput]):
    r"""RNN decoder with attention mechanism.

    Args:
        input_size (int): Dimension of input embeddings.
        encoder_output_size (int): The output size of the encoder cell.
        vocab_size (int): Vocabulary size. Required if
            :attr:`output_layer` is `None`.
        token_embedder: An instance of :torch_nn:`Module`, or a function taking
            a :tensor:`LongTensor` ``tokens`` as argument. This is the embedder
            called in :meth:`embed_tokens` to convert input tokens to
            embeddings.
        token_pos_embedder: An instance of :torch_nn:`Module`, or a function
            taking two :tensor:`LongTensor`\ s ``tokens`` and ``positions`` as
            argument. This is the embedder called in :meth:`embed_tokens` to
            convert input tokens with positions to embeddings.

            .. note::
                Only one among :attr:`token_embedder` and
                :attr:`token_pos_embedder` should be specified. If neither is
                specified, you must subclass :class:`AttentionRNNDecoder` and
                override :meth:`embed_tokens`.
        cell (RNNCellBase, optional): An instance of
            :class:`~texar.torch.core.cell_wrappers.RNNCellBase`. If `None`,
            a cell is created as specified in :attr:`hparams`.
        output_layer (optional): An output layer that transforms cell output
            to logits. This can be:

            - A callable layer, e.g., an instance of :torch_nn:`Module`.
            - A tensor. A dense layer will be created using the tensor
              as the kernel weights. The bias of the dense layer is determined
              by `hparams.output_layer_bias`. This can be used to tie the
              output layer with the input embedding matrix, as proposed in
              https://arxiv.org/pdf/1608.05859.pdf
            - `None`. A dense layer will be created based on :attr:`vocab_size`
              and `hparams.output_layer_bias`.
            - If no output layer after the cell output is needed, set
              `(vocab_size=None, output_layer=texar.torch.core.identity)`.
        cell_input_fn (callable, optional): A callable that produces RNN cell
            inputs. If `None` (default), the default is used:
            :python:`lambda inputs, attention:
            torch.cat([inputs, attention], -1)`,
            which concatenates regular RNN cell inputs with attentions.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    See :meth:`texar.torch.modules.RNNDecoderBase.forward` for the inputs and
    outputs of the decoder. The decoder returns
    `(outputs, final_state, sequence_lengths)`, where `outputs` is an instance
    of :class:`~texar.torch.modules.AttentionRNNDecoderOutput`.

    Example:
        .. code-block:: python

            # Encodes the source
            enc_embedder = WordEmbedder(data.source_vocab.size, ...)
            encoder = UnidirectionalRNNEncoder(...)
            enc_outputs, _ = encoder(
                inputs=enc_embedder(data_batch['source_text_ids']),
                sequence_length=data_batch['source_length'])
            # Decodes while attending to the source
            dec_embedder = WordEmbedder(vocab_size=data.target_vocab.size, ...)
            decoder = AttentionRNNDecoder(
                encoder_output_size=(self.encoder.cell_fw.hidden_size +
                                     self.encoder.cell_bw.hidden_size),
                input_size=dec_embedder.dim,
                vocab_size=data.target_vocab.size)
            outputs, _, _ = decoder(
                decoding_strategy='train_greedy',
                memory=enc_outputs,
                memory_sequence_length=data_batch['source_length'],
                inputs=dec_embedder(data_batch['target_text_ids']),
                sequence_length=data_batch['target_length']-1)
    """

    def __init__(self,
                 input_size: int,
                 encoder_output_size: int,
                 vocab_size: int,
                 token_embedder: Optional[TokenEmbedder] = None,
                 token_pos_embedder: Optional[TokenPosEmbedder] = None,
                 cell: Optional[RNNCellBase] = None,
                 output_layer: Optional[Union[nn.Module, torch.Tensor]] = None,
                 cell_input_fn: Optional[Callable[[torch.Tensor, torch.Tensor],
                                                  torch.Tensor]] = None,
                 hparams=None):

        super().__init__(
            input_size, vocab_size, token_embedder, token_pos_embedder,
            cell=cell, output_layer=output_layer, hparams=hparams)

        attn_hparams = self._hparams['attention']
        attn_kwargs = attn_hparams['kwargs'].todict()

        # Compute the correct input_size internally.
        if cell is None:
            if cell_input_fn is None:
                if attn_hparams["attention_layer_size"] is None:
                    input_size += encoder_output_size
                else:
                    input_size += attn_hparams["attention_layer_size"]
            else:
                if attn_hparams["attention_layer_size"] is None:
                    input_size = cell_input_fn(
                        torch.empty(input_size),
                        torch.empty(encoder_output_size)).shape[-1]
                else:
                    input_size = cell_input_fn(
                        torch.empty(input_size),
                        torch.empty(
                            attn_hparams["attention_layer_size"])).shape[-1]
            self._cell = layers.get_rnn_cell(input_size, self._hparams.rnn_cell)

        # Parse the `probability_fn` argument
        if 'probability_fn' in attn_kwargs:
            prob_fn = attn_kwargs['probability_fn']
            if prob_fn is not None and not callable(prob_fn):
                prob_fn = get_function(prob_fn, ['torch.nn.functional',
                                                 'texar.torch.core'])
            attn_kwargs['probability_fn'] = prob_fn

        # Parse `encoder_output_size` and `decoder_output_size` arguments
        if attn_hparams['type'] in ['BahdanauAttention',
                                    'BahdanauMonotonicAttention']:
            attn_kwargs.update({"decoder_output_size": self._cell.hidden_size})
        attn_kwargs.update({"encoder_output_size": encoder_output_size})

        attn_modules = ['texar.torch.core']
        # TODO: Support multiple attention mechanisms.
        self.attention_mechanism: AttentionMechanism
        self.attention_mechanism = check_or_get_instance(
            attn_hparams["type"], attn_kwargs, attn_modules,
            classtype=AttentionMechanism)

        self._attn_cell_kwargs = {
            "attention_layer_size": attn_hparams["attention_layer_size"],
            "alignment_history": attn_hparams["alignment_history"],
            "output_attention": attn_hparams["output_attention"],
        }
        self._cell_input_fn = cell_input_fn

        if attn_hparams["output_attention"] and vocab_size is not None and \
                self.attention_mechanism is not None:
            if attn_hparams["attention_layer_size"] is None:
                self._output_layer = nn.Linear(
                    encoder_output_size,
                    vocab_size)
            else:
                self._output_layer = nn.Linear(
                    sum(attn_hparams["attention_layer_size"])
                    if isinstance(attn_hparams["attention_layer_size"], list)
                    else attn_hparams["attention_layer_size"],
                    vocab_size)

        attn_cell = AttentionWrapper(
            self._cell,
            self.attention_mechanism,
            cell_input_fn=self._cell_input_fn,
            **self._attn_cell_kwargs)

        self._cell: AttentionWrapper = attn_cell

        self.memory: Optional[torch.Tensor] = None
        self.memory_sequence_length: Optional[torch.LongTensor] = None

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.
        Common hyperparameters are the same as in
        :class:`~texar.torch.modules.BasicRNNDecoder`.
        :meth:`~texar.torch.modules.BasicRNNDecoder.default_hparams`.
        Additional hyperparameters are for attention mechanism
        configuration.

        .. code-block:: python

            {
                "attention": {
                    "type": "LuongAttention",
                    "kwargs": {
                        "num_units": 256,
                    },
                    "attention_layer_size": None,
                    "alignment_history": False,
                    "output_attention": True,
                },
                # The following hyperparameters are the same as with
                # `BasicRNNDecoder`
                "rnn_cell": default_rnn_cell_hparams(),
                "max_decoding_length_train": None,
                "max_decoding_length_infer": None,
                "helper_train": {
                    "type": "TrainingHelper",
                    "kwargs": {}
                }
                "helper_infer": {
                    "type": "SampleEmbeddingHelper",
                    "kwargs": {}
                }
                "name": "attention_rnn_decoder"
            }

        Here:

        `"attention"`: dict
            Attention hyperparameters, including:

            `"type"`: str or class or instance
                The attention type. Can be an attention class, its name or
                module path, or a class instance. The class must be a subclass
                of ``AttentionMechanism``. See :ref:`attention-mechanism` for
                all supported attention mechanisms. If class name is given,
                the class must be from modules
                :mod:`texar.torch.core` or :mod:`texar.torch.custom`.

                Example:

                .. code-block:: python

                    # class name
                    "type": "LuongAttention"
                    "type": "BahdanauAttention"
                    # module path
                    "type": "texar.torch.core.BahdanauMonotonicAttention"
                    "type": "my_module.MyAttentionMechanismClass"
                    # class
                    "type": texar.torch.core.LuongMonotonicAttention
                    # instance
                    "type": LuongAttention(...)

            `"kwargs"`: dict
                keyword arguments for the attention class constructor.
                Arguments :attr:`memory` and
                :attr:`memory_sequence_length` should **not** be
                specified here because they are given to the decoder
                constructor. Ignored if "type" is an attention class
                instance. For example:

                .. code-block:: python

                    "type": "LuongAttention",
                    "kwargs": {
                        "num_units": 256,
                        "probability_fn": torch.nn.functional.softmax,
                    }

                Here `"probability_fn"` can also be set to the string name
                or module path to a probability function.

                `"attention_layer_size"`: int or None
                    The depth of the attention (output) layer. The context and
                    cell output are fed into the attention layer to generate
                    attention at each time step.
                    If `None` (default), use the context as attention at each
                    time step.

                `"alignment_history"`: bool
                    whether to store alignment history from all time steps
                    in the final output state. (Stored as a time major
                    `TensorArray` on which you must call `stack()`.)

                `"output_attention"`: bool
                    If `True` (default), the output at each time step is
                    the attention value. This is the behavior of Luong-style
                    attention mechanisms. If `False`, the output at each
                    time step is the output of `cell`.  This is the
                    behavior of Bahdanau-style attention mechanisms.
                    In both cases, the `attention` tensor is propagated to
                    the next time step via the state and is used there.
                    This flag only controls whether the attention mechanism
                    is propagated up to the next cell in an RNN stack or to
                    the top RNN output.
        """
        hparams = RNNDecoderBase.default_hparams()
        hparams["name"] = "attention_rnn_decoder"
        hparams["attention"] = {
            "type": "LuongAttention",
            "kwargs": {
                "num_units": 256,
            },
            "attention_layer_size": None,
            "alignment_history": False,
            "output_attention": True,
        }
        return hparams

    def initialize(  # type: ignore
            self, helper: Helper,
            inputs: Optional[torch.Tensor],
            sequence_length: Optional[torch.LongTensor],
            initial_state: Optional[MaybeList[MaybeTuple[torch.Tensor]]]) -> \
            Tuple[torch.ByteTensor, torch.Tensor,
                  Optional[AttentionWrapperState]]:
        initial_finished, initial_inputs = helper.initialize(
            self.embed_tokens, inputs, sequence_length)
        if initial_state is None:
            state = None
        else:
            tensor = utils.get_first_in_structure(initial_state)
            assert tensor is not None
            tensor: torch.Tensor
            state = self._cell.zero_state(batch_size=tensor.size(0))
            state = state._replace(cell_state=initial_state)

        return initial_finished, initial_inputs, state

    def step(self, helper: Helper, time: int, inputs: torch.Tensor,
             state: Optional[AttentionWrapperState]) -> \
            Tuple[AttentionRNNDecoderOutput, AttentionWrapperState]:
        wrapper_outputs, wrapper_state = self._cell(
            inputs, state, self.memory, self.memory_sequence_length)
        # Essentially the same as in BasicRNNDecoder.step()

        logits = self._output_layer(wrapper_outputs)
        sample_ids = helper.sample(time=time, outputs=logits)

        attention_scores = wrapper_state.alignments
        attention_context = wrapper_state.attention
        outputs = AttentionRNNDecoderOutput(
            logits, sample_ids, wrapper_outputs,
            attention_scores, attention_context)
        next_state = wrapper_state

        return outputs, next_state

    def next_inputs(self, helper: Helper, time: int,
                    outputs: AttentionRNNDecoderOutput) -> \
            Tuple[torch.Tensor, torch.ByteTensor]:
        finished, next_inputs = helper.next_inputs(
            self.embed_tokens, time, outputs.logits, outputs.sample_id)
        return next_inputs, finished

    def forward(  # type: ignore
            self,
            memory: torch.Tensor,
            memory_sequence_length: Optional[torch.LongTensor] = None,
            inputs: Optional[torch.Tensor] = None,
            sequence_length: Optional[torch.LongTensor] = None,
            initial_state: Optional[MaybeList[MaybeTuple[torch.Tensor]]] = None,
            helper: Optional[Helper] = None,
            max_decoding_length: Optional[int] = None,
            impute_finished: bool = False,
            infer_mode: Optional[bool] = None,
            beam_width: Optional[int] = None,
            length_penalty: float = 0., **kwargs) \
            -> Union[Tuple[AttentionRNNDecoderOutput,
                           Optional[AttentionWrapperState], torch.LongTensor],
                     Dict[str, torch.Tensor]]:
        r"""Performs decoding.

        Implementation calls initialize() once and step() repeatedly on the
        Decoder object. Please refer to `tf.contrib.seq2seq.dynamic_decode`.

        See Also:
            Arguments of :meth:`create_helper`.

        Args:
            memory: The memory to query; usually the output of an RNN encoder.
                This tensor should be shaped `[batch_size, max_time, ...]`.
            memory_sequence_length: (optional) Sequence lengths for the batch
                entries in memory.  If provided, the memory tensor rows are
                masked with zeros for values past the respective sequence
                lengths.
            inputs (optional): Input tensors for teacher forcing decoding.
                Used when :attr:`decoding_strategy` is set to
                ``"train_greedy"``, or when `hparams`-configured helper is used.

                The attr:`inputs` is a :tensor:`LongTensor` used as index to
                look up embeddings and feed in the decoder. For example, if
                :attr:`embedder` is an instance of
                :class:`~texar.torch.modules.WordEmbedder`, then :attr:`inputs`
                is usually a 2D int Tensor `[batch_size, max_time]` (or
                `[max_time, batch_size]` if `input_time_major` == `True`)
                containing the token indexes.
            sequence_length (optional): A 1D int Tensor containing the
                sequence length of :attr:`inputs`.
                Used when `decoding_strategy="train_greedy"` or
                `hparams`-configured helper is used.
            initial_state (optional): Initial state of decoding.
                If `None` (default), zero state is used.
            helper (optional): An instance of
                :class:`~texar.torch.modules.Helper`
                that defines the decoding strategy. If given,
                ``decoding_strategy`` and helper configurations in
                :attr:`hparams` are ignored.
            max_decoding_length: A int scalar Tensor indicating the maximum
                allowed number of decoding steps. If `None` (default), either
                `hparams["max_decoding_length_train"]` or
                `hparams["max_decoding_length_infer"]` is used
                according to :attr:`mode`.
            impute_finished (bool): If `True`, then states for batch
                entries which are marked as finished get copied through and
                the corresponding outputs get zeroed out.  This causes some
                slowdown at each time step, but ensures that the final state
                and outputs have the correct values and that backprop ignores
                time steps that were marked as finished.
            infer_mode (optional): If not `None`, overrides mode given by
                `self.training`.
            beam_width (int): Set to use beam search. If given,
                :attr:`decoding_strategy` is ignored.
            length_penalty (float): Length penalty coefficient used in beam
                search decoding. Refer to https://arxiv.org/abs/1609.08144
                for more details.
                It should be larger if longer sentences are desired.
            **kwargs: Other keyword arguments for constructing helpers
                defined by ``hparams["helper_train"]`` or
                ``hparams["helper_infer"]``.

        Returns:

            - For **beam search** decoding, returns a ``dict`` containing keys
              ``"sample_id"`` and ``"log_prob"``.

                - ``"sample_id"`` is a :tensor:`LongTensor` of shape
                  ``[batch_size, max_time, beam_width]`` containing generated
                  token indexes. ``sample_id[:,:,0]`` is the highest-probable
                  sample.
                - ``"log_prob"`` is a :tensor:`Tensor` of shape
                  ``[batch_size, beam_width]`` containing the log probability
                  of each sequence sample.

            - For **"infer_greedy"** and **"infer_sample"** decoding or
              decoding with :attr:`helper`, returns
              a tuple `(outputs, final_state, sequence_lengths)`, where

                - **outputs**: an object containing the decoder output on all
                  time steps.
                - **final_state**: is the cell state of the final time step.
                - **sequence_lengths**: is an int Tensor of shape `[batch_size]`
                  containing the length of each sample.
        """
        # TODO: Add faster code path for teacher-forcing training.

        # Save memory and memory_sequence_length
        self.memory = memory
        self.memory_sequence_length = memory_sequence_length

        # Maximum decoding length
        if max_decoding_length is None:
            if self.training:
                max_decoding_length = self._hparams.max_decoding_length_train
            else:
                max_decoding_length = self._hparams.max_decoding_length_infer
            if max_decoding_length is None:
                max_decoding_length = utils.MAX_SEQ_LENGTH

        # Beam search decode
        if beam_width is not None and beam_width > 1:
            if helper is not None:
                raise ValueError("Must not set 'beam_width' and 'helper' "
                                 "simultaneously.")

            start_tokens = kwargs.get('start_tokens')
            if start_tokens is None:
                raise ValueError("'start_tokens' must be specified when using"
                                 "beam search decoding.")

            batch_size = start_tokens.shape[0]
            state = self._cell.zero_state(batch_size)
            if initial_state is not None:
                state = state._replace(cell_state=initial_state)

            end_token = kwargs.get('end_token')
            assert isinstance(end_token, int)

            sample_id, log_prob = self.beam_decode(
                start_tokens=start_tokens,
                end_token=end_token,
                initial_state=state,
                beam_width=beam_width,
                length_penalty=length_penalty,
                decode_length=max_decoding_length)

            # Release memory and memory_sequence_length in AttentionRNNDecoder
            self.memory = None
            self.memory_sequence_length = None

            # Release the cached memory in AttentionMechanism
            for attention_mechanism in self._cell.attention_mechanisms:
                attention_mechanism.clear_cache()

            return {"sample_id": sample_id,
                    "log_prob": log_prob}

        # Helper
        if helper is None:
            helper = self._create_or_get_helper(infer_mode, **kwargs)

        if (isinstance(helper, decoder_helpers.TrainingHelper) and
                (inputs is None or sequence_length is None)):
            raise ValueError("'input' and 'sequence_length' must not be None "
                             "when using 'TrainingHelper'.")

        # Initial state
        self._cell.init_batch()

        (outputs, final_state,
         sequence_lengths) = self.dynamic_decode(
            helper, inputs, sequence_length, initial_state,  # type: ignore
            max_decoding_length, impute_finished)

        # Release memory and memory_sequence_length in AttentionRNNDecoder
        self.memory = None
        self.memory_sequence_length = None

        # Release the cached memory in AttentionMechanism
        for attention_mechanism in self._cell.attention_mechanisms:
            attention_mechanism.clear_cache()

        return outputs, final_state, sequence_lengths

    def beam_decode(self,
                    start_tokens: torch.LongTensor,
                    end_token: int,
                    initial_state: AttentionWrapperState,
                    decode_length: int = 256,
                    beam_width: int = 5,
                    length_penalty: float = 0.6) \
            -> Tuple[torch.LongTensor, torch.Tensor]:

        def _prepare_beam_search(x):
            x = x.unsqueeze(1).repeat(1, beam_width, *([1] * (x.dim() - 1)))
            x = x.view(-1, *x.size()[2:])
            return x

        memory_beam_search = _prepare_beam_search(self.memory)
        memory_sequence_length_beam_search = _prepare_beam_search(
            self.memory_sequence_length)

        def _symbols_to_logits_fn(ids, state):
            batch_size = ids.size(0)
            step = ids.size(-1) - 1
            times = ids.new_full((batch_size,), step)
            inputs = self.embed_tokens(ids[:, -1], times)
            wrapper_outputs, wrapper_state = self._cell(
                inputs, state, memory_beam_search,
                memory_sequence_length_beam_search)
            logits = self._output_layer(wrapper_outputs)
            return logits, wrapper_state

        assert self._vocab_size is not None
        outputs, log_prob = beam_search(
            symbols_to_logits_fn=_symbols_to_logits_fn,
            initial_ids=start_tokens,
            beam_size=beam_width,
            decode_length=decode_length,
            vocab_size=self._vocab_size,
            alpha=length_penalty,
            states=initial_state,
            eos_id=end_token)

        # Ignores <BOS>
        outputs = outputs[:, :, 1:]
        # shape = [batch_size, seq_length, beam_width]
        outputs = outputs.permute((0, 2, 1))
        return outputs, log_prob

    @property
    def output_size(self):
        r"""Output size of one step.
        """
        return self._cell.output_size



# Copyright 2018 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Helper functions and classes for embedding processing.
"""
from typing import Callable, Dict

import numpy as np

from texar.torch.hyperparams import HParams
from texar.torch.utils import utils

__all__ = [
    "load_word2vec",
    "load_glove",
    "Embedding",
]


def load_word2vec(filename: str, vocab: Dict[str, int],
                  word_vecs: np.ndarray) -> np.ndarray:
    r"""Loads embeddings in the word2vec binary format which has a header line
    containing the number of vectors and their dimensionality (two integers),
    followed with number-of-vectors lines each of which is formatted as
    ``<word-string> <embedding-vector>``.

    Args:
        filename (str): Path to the embedding file.
        vocab (dict): A dictionary that maps token strings to integer index.
            Tokens not in :attr:`vocab` are not read.
        word_vecs: A 2D numpy array of shape `[vocab_size, embed_dim]`
            which is updated as reading from the file.

    Returns:
        The updated :attr:`word_vecs`.
    """
    with open(filename, "rb") as fin:
        header = fin.readline()
        vocab_size, vector_size = [int(s) for s in header.split()]
        if vector_size != word_vecs.shape[1]:
            raise ValueError("Inconsistent word vector sizes: %d vs %d" %
                             (vector_size, word_vecs.shape[1]))
        binary_len = np.dtype('float32').itemsize * vector_size
        for _ in np.arange(vocab_size):
            chars = []
            while True:
                char = fin.read(1)
                if char == b' ':
                    break
                if char != b'\n':
                    chars.append(char)
            word = b''.join(chars).decode('utf-8')
            if word in vocab:
                word_vecs[vocab[word]] = np.frombuffer(
                    fin.read(binary_len), dtype='float32')
            else:
                fin.read(binary_len)
    return word_vecs


def load_glove(filename: str, vocab: Dict[str, int],
               word_vecs: np.ndarray) -> np.ndarray:
    r"""Loads embeddings in the glove text format in which each line is
    ``<word-string> <embedding-vector>``. Dimensions of the embedding vector
    are separated with whitespace characters.

    Args:
        filename (str): Path to the embedding file.
        vocab (dict): A dictionary that maps token strings to integer index.
            Tokens not in :attr:`vocab` are not read.
        word_vecs: A 2D numpy array of shape `[vocab_size, embed_dim]`
            which is updated as reading from the file.

    Returns:
        The updated :attr:`word_vecs`.
    """
    with open(filename) as fin:
        for line in fin:
            vec = line.strip().split()
            if len(vec) == 0:
                continue
            word, vec = vec[0], vec[1:]
            if word not in vocab:
                continue
            if len(vec) != word_vecs.shape[1]:
                raise ValueError("Inconsistent word vector sizes: %d vs %d" %
                                 (len(vec), word_vecs.shape[1]))
            word_vecs[vocab[word]] = np.array([float(v) for v in vec])
    return word_vecs


class Embedding:
    r"""Embedding class that loads token embedding vectors from file. Token
    embeddings not in the embedding file are initialized as specified in
    :attr:`hparams`.

    Args:
        vocab (dict): A dictionary that maps token strings to integer index.
        hparams (dict): Hyperparameters. See :meth:`default_hparams` for the
            defaults.
    """

    def __init__(self, vocab: Dict[str, int],
                 hparams=None):
        self._hparams = HParams(hparams, self.default_hparams())

        # Initialize embeddings
        init_fn_kwargs = self._hparams.init_fn.kwargs.todict()
        if "shape" in init_fn_kwargs or "size" in init_fn_kwargs:
            raise ValueError("Argument 'shape' or 'size' must not be "
                             "specified. They are inferred automatically.")
        init_fn: Callable[..., np.ndarray]
        init_fn = utils.get_function(
            self._hparams.init_fn.type,
            ["numpy.random", "numpy", "texar.torch.custom"])

        try:
            self._word_vecs = init_fn(  # type: ignore
                size=[len(vocab), self._hparams.dim], **init_fn_kwargs)
        except TypeError:
            self._word_vecs = init_fn(  # type: ignore
                shape=[len(vocab), self._hparams.dim], **init_fn_kwargs)

        # Optionally read embeddings from file
        if self._hparams.file is not None and self._hparams.file != "":
            read_fn: Callable[[str, Dict[str, int], np.ndarray], np.ndarray]
            read_fn = utils.get_function(  # type: ignore
                self._hparams.read_fn,
                ["texar.torch.data.embedding", "texar.torch.data",
                 "texar.torch.custom"])

            self._word_vecs = read_fn(self._hparams.file,
                                      vocab, self._word_vecs)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values:

        .. code-block:: python

            {
                "file": "",
                "dim": 50,
                "read_fn": "load_word2vec",
                "init_fn": {
                    "type": "numpy.random.uniform",
                    "kwargs": {
                        "low": -0.1,
                        "high": 0.1,
                    }
                },
            }

        Here:

        `"file"`: str
            Path to the embedding file. If not provided, all embeddings are
            initialized with the initialization function.

        `"dim"`: int
            Dimension size of each embedding vector

        `"read_fn"`: str or callable
            Function to read the embedding file. This can be the function,
            or its string name or full module path. For example,

            .. code-block:: python

                "read_fn": texar.torch.data.load_word2vec
                "read_fn": "load_word2vec"
                "read_fn": "texar.torch.data.load_word2vec"
                "read_fn": "my_module.my_read_fn"

            If function string name is used, the function must be in
            one of the modules: :mod:`texar.torch.data` or
            :mod:`texar.torch.custom`.

            The function must have the same signature as with
            :func:`load_word2vec`.

        `"init_fn"`: dict
            Hyperparameters of the initialization function used to initialize
            embedding of tokens missing in the embedding
            file.

            The function must accept argument named `size` or `shape` to
            specify the output shape, and return a numpy array of the shape.

            The `dict` has the following fields:

            `"type"`: str or callable
                The initialization function. Can be either the function,
                or its string name or full module path.

            `"kwargs"`: dict
                Keyword arguments for calling the function. The function
                is called with :python:`init_fn(size=[.., ..], **kwargs)`.
        """
        return {
            "file": "",
            "dim": 50,
            "read_fn": "load_word2vec",
            "init_fn": {
                "type": "numpy.random.uniform",
                "kwargs": {
                    "low": -0.1,
                    "high": 0.1,
                },
            },
            "@no_typecheck": ["read_fn", "init_fn"]
        }

    @property
    def word_vecs(self):
        r"""2D numpy array of shape `[vocab_size, embedding_dim]`.
        """
        return self._word_vecs

    @property
    def vector_size(self):
        r"""The embedding dimension size.
        """
        return self._hparams.dim



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for connectors.
"""

from __future__ import unicode_literals

from collections import namedtuple
import unittest
import torch
import torch.distributions as tds

from texar.torch.core import layers
from texar.torch.modules.connectors.connectors import ConstantConnector
from texar.torch.modules.connectors.connectors import MLPTransformConnector
from texar.torch.modules.connectors.connectors import (
    ReparameterizedStochasticConnector)
from texar.torch.modules.connectors.connectors import _assert_same_size

from texar.torch.utils import nest


class TestConnectors(unittest.TestCase):
    r"""Tests various connectors.
    """

    def setUp(self) -> None:
        self._batch_size = 100

        self._decoder_cell = layers.get_rnn_cell(
            256, layers.default_rnn_cell_hparams())

    def test_constant_connector(self):
        r"""Tests the logic of
        :class:`~texar.torch.modules.connectors.ConstantConnector`.
        """

        state_size = namedtuple('LSTMStateTuple', ['h', 'c'])(256, 256)
        connector_0 = ConstantConnector(state_size)
        decoder_initial_state_0 = connector_0(self._batch_size)
        connector_1 = ConstantConnector(
            state_size, hparams={"value": 1.})
        decoder_initial_state_1 = connector_1(self._batch_size)

        s_0 = decoder_initial_state_0
        s_1 = decoder_initial_state_1
        self.assertEqual(nest.flatten(s_0)[0][0, 0], 0.)
        self.assertEqual(nest.flatten(s_1)[0][0, 0], 1.)

        size = torch.Size([1, 2, 3])
        connector_size_0 = ConstantConnector(
            size, hparams={"value": 2.})
        size_tensor = connector_size_0(self._batch_size)
        self.assertEqual(
            torch.Size([self._batch_size]) + size, size_tensor.size())
        self.assertEqual(size_tensor[0][0, 0, 0], 2.)

        tuple_size_1 = (torch.Size([1, 2, 3]), torch.Size([4, 5, 6]))
        connector_size_1 = ConstantConnector(
            tuple_size_1, hparams={"value": 3.})
        tuple_size_tensor = connector_size_1(self._batch_size)
        tuple_size_tensor_0 = tuple_size_tensor[0]
        tuple_size_tensor_1 = tuple_size_tensor[1]
        self.assertEqual(
            torch.Size([self._batch_size]) + torch.Size([1, 2, 3]),
            tuple_size_tensor_0.size())
        self.assertEqual(tuple_size_tensor_0[0][0, 0, 0], 3.)
        self.assertEqual(
            torch.Size([self._batch_size]) + torch.Size([4, 5, 6]),
            tuple_size_tensor_1.size())
        self.assertEqual(tuple_size_tensor_1[0][0, 0, 0], 3.)

        tuple_size_2 = (5, 10)
        connector_size_2 = ConstantConnector(
            tuple_size_2, hparams={"value": 4.})
        tuple_size_tensor = connector_size_2(self._batch_size)
        tuple_size_tensor_0 = tuple_size_tensor[0]
        tuple_size_tensor_1 = tuple_size_tensor[1]
        self.assertEqual(
            torch.Size([self._batch_size]) + torch.Size([5]),
            tuple_size_tensor_0.size())
        self.assertEqual(tuple_size_tensor_0[0][0], 4.)
        self.assertEqual(
            torch.Size([self._batch_size]) + torch.Size([10]),
            tuple_size_tensor_1.size())
        self.assertEqual(tuple_size_tensor_1[0][0], 4.)

        tuple_size_3 = (torch.Size([1, 2, 3]), 10)
        connector_size_3 = ConstantConnector(
            tuple_size_3, hparams={"value": 4.})
        tuple_size_tensor = connector_size_3(self._batch_size)
        tuple_size_tensor_0 = tuple_size_tensor[0]
        tuple_size_tensor_1 = tuple_size_tensor[1]
        self.assertEqual(
            torch.Size([self._batch_size]) + torch.Size([1, 2, 3]),
            tuple_size_tensor_0.size())
        self.assertEqual(tuple_size_tensor_0[0][0, 0, 0], 4.)
        self.assertEqual(
            torch.Size([self._batch_size]) + torch.Size([10]),
            tuple_size_tensor_1.size())
        self.assertEqual(tuple_size_tensor_1[0][0], 4.)

    def test_forward_connector(self):
        r"""Tests the logic of
        :class:`~texar.torch.modules.connectors.ForwardConnector`.
        """
        # TODO(zhiting)
        pass

    def test_mlp_transform_connector(self):
        r"""Tests the logic of
        :class:`~texar.torch.modules.connectors.MLPTransformConnector`.
        """
        state_size = namedtuple('LSTMStateTuple', ['h', 'c'])(256, 256)
        connector = MLPTransformConnector(state_size, linear_layer_dim=10)
        output = connector(torch.zeros(5, 10))
        output_1 = output[0]
        output_2 = output[1]
        self.assertEqual(
            torch.Size([5, 256]),
            output_1.size())
        self.assertEqual(
            torch.Size([5, 256]),
            output_2.size())

        state_size = (16, 32)
        connector = MLPTransformConnector(state_size, linear_layer_dim=10)
        output = connector(torch.zeros(5, 10))
        output_1 = output[0]
        output_2 = output[1]
        self.assertEqual(
            torch.Size([5, 16]),
            output_1.size())
        self.assertEqual(
            torch.Size([5, 32]),
            output_2.size())

        state_size = (torch.Size([8, 32]), torch.Size([16, 64]))
        connector = MLPTransformConnector(state_size, linear_layer_dim=10)
        output = connector(torch.zeros(5, 10))
        output_1 = output[0]
        output_2 = output[1]
        self.assertEqual(
            torch.Size([5, 8, 32]),
            output_1.size())
        self.assertEqual(
            torch.Size([5, 16, 64]),
            output_2.size())

    def test_reparameterized_stochastic_connector(self):
        r"""Tests the logic of
        :class:`~texar.torch.modules.ReparameterizedStochasticConnector`.
        """

        self._batch_size = 16
        variable_size = 100
        sample_num = 10

        mu = torch.zeros([self._batch_size, variable_size])
        var = torch.ones([variable_size])

        state_size = (10, 11)
        gauss_connector = ReparameterizedStochasticConnector(
            state_size,
            mlp_input_size=mu.size()[-1:],
            distribution="MultivariateNormal",
            distribution_kwargs={"loc": mu, "scale_tril": torch.diag(var)})
        output, samples = gauss_connector()

        self.assertEqual(output[0].size(),
                         torch.Size([self._batch_size, state_size[0]]))
        self.assertEqual(output[1].size(),
                         torch.Size([self._batch_size, state_size[1]]))
        self.assertEqual(samples.size(),
                         torch.Size([self._batch_size, variable_size]))

        output, _ = gauss_connector(num_samples=12, transform=False)
        self.assertEqual(output.size(),
                         torch.Size([12, self._batch_size, variable_size]))

        state_size_ts = (torch.Size([10, 10]), torch.Size([2, 3, 4]))
        gauss_connector_ts = ReparameterizedStochasticConnector(
            state_size_ts,
            mlp_input_size=mu.size()[-1:],
            distribution="MultivariateNormal",
            distribution_kwargs={"loc": mu, "scale_tril": torch.diag(var)})
        output, samples = gauss_connector_ts()

        _assert_same_size(output, state_size_ts)

        # sample_mu = np.mean(sample_outputs, axis=0)
        # # pylint: disable=no-member
        # sample_var = np.var(sample_outputs, axis=0)

        # # check if the value is approximated N(0, 1)
        # for i in range(variable_size):
        #     self.assertAlmostEqual(0, sample_mu[i], delta=0.2)
        #     self.assertAlmostEqual(1, sample_var[i], delta=0.2)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Transformer encoders with multi-head self attention.
"""

from typing import Any, Dict

import torch
from torch import nn

from texar.torch.core import layers
from texar.torch.modules.encoders.encoder_base import EncoderBase
from texar.torch.modules.encoders.multihead_attention import (
    MultiheadAttentionEncoder)
from texar.torch.modules.networks.networks import FeedForwardNetwork
from texar.torch.utils import transformer_attentions as attn
from texar.torch.utils.utils import sequence_mask

__all__ = [
    "default_transformer_poswise_net_hparams",
    "TransformerEncoder",
]


def default_transformer_poswise_net_hparams(input_dim: int,
                                            output_dim: int = 512) \
        -> Dict[str, Any]:
    r"""Returns default hyperparameters of a
    :class:`~texar.torch.modules.FeedForwardNetwork` as a position-wise network
    used in :class:`~texar.torch.modules.TransformerEncoder` and
    :class:`~texar.torch.modules.TransformerDecoder`.
    This is a 2-layer dense network with dropout in-between.

    .. code-block:: python

        {
            "layers": [
                {
                    "type": "Linear",
                    "kwargs": {
                        "in_features": input_dim,
                        "out_features": output_dim * 4,
                        "bias": True,
                    }
                },
                {
                    "type": "nn.ReLU",
                    "kwargs": {
                        "inplace": True
                    }
                },
                {
                    "type": "Dropout",
                    "kwargs": {
                        "p": 0.1,
                    }
                },
                {
                    "type": "Linear",
                    "kwargs": {
                        "in_features": output_dim * 4,
                        "out_features": output_dim,
                        "bias": True,
                    }
                }
            ],
            "name": "ffn"
        }

    Args:
        input_dim (int): The size of dense layer input.
        output_dim (int): The size of dense layer output.
    """
    return {
        "layers": [
            {
                "type": "Linear",
                "kwargs": {
                    "in_features": input_dim,
                    "out_features": output_dim * 4,
                    "bias": True,
                }
            },
            {
                "type": "ReLU",
                "kwargs": {
                    "inplace": True
                }
            },
            {
                "type": "Dropout",
                "kwargs": {
                    "p": 0.1,
                }
            },
            {
                "type": "Linear",
                "kwargs": {
                    "in_features": output_dim * 4,
                    "out_features": output_dim,
                    "bias": True,
                }
            }
        ],
        "name": "ffn"
    }


class TransformerEncoder(EncoderBase):
    r"""Transformer encoder that applies multi-head self attention for encoding
    sequences.

    This module basically stacks
    :class:`~texar.torch.modules.MultiheadAttentionEncoder`,
    :class:`~texar.torch.modules.FeedForwardNetwork` and residual connections.
    This module supports two types of architectures, namely, the standard
    Transformer Encoder architecture first proposed in
    `(Vaswani et al.) "Attention is All You Need"`, and
    the variant first used in `(Devlin et al.)` BERT. See
    :meth:`default_hparams` for the nuance between the two types of
    architectures.

    Args:
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    .. document private functions
    """

    def __init__(self, hparams=None):
        super().__init__(hparams=hparams)
        self._input_size = self._hparams.dim
        self.self_attns = nn.ModuleList()
        if not self._hparams.use_bert_config:
            self.self_attn_layer_norm = nn.ModuleList()
        else:
            self.output_layer_norm = nn.ModuleList()
        self.poswise_networks = nn.ModuleList()
        self.poswise_layer_norm = nn.ModuleList()

        # In PyTorch, eps for LayerNorm is 1e-6 by default.
        eps = self._hparams.eps

        self.initialize_blocks()

        self.embed_dropout = nn.Dropout(p=self._hparams.embedding_dropout)
        self.residual_dropout = nn.Dropout(p=self._hparams.residual_dropout)

        if self._hparams.use_bert_config:
            self.input_normalizer = nn.LayerNorm(self._input_size, eps=eps)
        else:
            self.final_layer_norm = nn.LayerNorm(self._input_size, eps=eps)

        if self._hparams.initializer:
            initialize = layers.get_initializer(self._hparams.initializer)
            assert initialize is not None
            # Do not re-initialize LayerNorm modules.
            for name, param in self.named_parameters():
                if name.split('.')[-1] == 'weight' and 'layer_norm' not in name:
                    initialize(param)

    def initialize_blocks(self):
        r"""Helper function which initializes blocks for encoder.

        Should be overridden by any classes where block initialization varies.
        """
        for _ in range(self._hparams.num_blocks):
            mh_attn = MultiheadAttentionEncoder(
                self._input_size, self._hparams.multihead_attention)
            self.self_attns.append(mh_attn)
            if not self._hparams.use_bert_config:
                self.self_attn_layer_norm.append(
                    nn.LayerNorm(self._input_size, eps=self._hparams.eps))
            if self._hparams.dim != mh_attn.hparams.output_dim:
                raise ValueError(
                    'The "dim" in the hparams of '
                    '"multihead_attention" should be equal to the '
                    '"dim" of TransformerEncoder')

            pw_net = FeedForwardNetwork(
                hparams=self._hparams['poswise_feedforward'])

            final_dim = pw_net.hparams.layers[-1]['kwargs']['out_features']
            if self._hparams.dim != final_dim:
                raise ValueError(
                    'The output dimenstion of '
                    '"poswise_feedforward" should be equal '
                    'to the "dim" of TransformerEncoder.')

            self.poswise_networks.append(pw_net)
            self.poswise_layer_norm.append(
                nn.LayerNorm(self._input_size, eps=self._hparams.eps))
            if self._hparams.use_bert_config:
                self.output_layer_norm.append(
                    nn.LayerNorm(self._input_size, eps=self._hparams.eps))

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "num_blocks": 6,
                "dim": 512,
                'use_bert_config': False,
                "embedding_dropout": 0.1,
                "residual_dropout": 0.1,
                "poswise_feedforward": default_transformer_poswise_net_hparams,
                'multihead_attention': {
                    'name': 'multihead_attention',
                    'num_units': 512,
                    'num_heads': 8,
                    'dropout_rate': 0.1,
                    'output_dim': 512,
                    'use_bias': False,
                },
                "eps": 1e-6,
                "initializer": None,
                "name": "transformer_encoder"
            }

        Here:

        `"num_blocks"`: int
            Number of stacked blocks.

        `"dim"`: int
            Hidden dimension of the encoders.

        `"use_bert_config"`: bool
            If `False`, apply the standard Transformer Encoder architecture from
            the original paper `(Vaswani et al.) "Attention is All You Need"`.
            If `True`, apply the Transformer Encoder architecture used in BERT
            `(Devlin et al.)` and the default setting of TensorFlow.
            The differences lie in:

            1. The standard arch restricts the word embedding of PAD token to
               all zero. The BERT arch does not.
            2. The attention bias for padding tokens: Standard architectures use
               ``-1e8`` for negative attention mask. BERT uses ``-1e4`` instead.
            3. The residual connections between internal tensors:
               In BERT, a residual layer connects the tensors *after* layer
               normalization. In standard architectures, the tensors are
               connected *before* layer normalization.

        `"embedding_dropout"`: float
            Dropout rate of the input embedding.

        `"residual_dropout"`: float
            Dropout rate of the residual connections.

        `"eps"`: float
            Epsilon values for layer norm layers.

        `"poswise_feedforward"`: dict
            Hyperparameters for a feed-forward network used in residual
            connections.
            Make sure the dimension of the output tensor is equal to ``"dim"``.
            See
            :func:`~texar.torch.modules.default_transformer_poswise_net_hparams`
            for details.

        `"multihead_attention"`: dict
            Hyperparameters for the multi-head attention strategy.
            Make sure the ``"output_dim"`` in this module is equal to ``"dim"``.
            See :class:`~texar.torch.modules.MultiheadAttentionEncoder` for
            details.

        `"initializer"`: dict, optional
            Hyperparameters of the default initializer that initializes
            variables created in this module.
            See :func:`~texar.torch.core.get_initializer` for details.

        `"name"`: str
            Name of the module.
        """
        dim = 512
        return {
            'num_blocks': 6,
            'dim': dim,
            'use_bert_config': False,
            'embedding_dropout': 0.1,
            'residual_dropout': 0.1,
            'poswise_feedforward': default_transformer_poswise_net_hparams(dim),
            'multihead_attention': {
                'name': 'multihead_attention',
                'num_units': 512,
                'num_heads': 8,
                'dropout_rate': 0.1,
                'output_dim': 512,
                'use_bias': False,
            },
            'initializer': None,
            'eps': 1e-6,
            'name': 'transformer_encoder',
        }

    def forward(self,  # type: ignore
                inputs: torch.Tensor,
                sequence_length: torch.LongTensor) -> torch.Tensor:
        r"""Encodes the inputs.

        Args:
            inputs: A 3D Tensor of shape ``[batch_size, max_time, dim]``,
                containing the embedding of input sequences. Note that
                the embedding dimension `dim` must equal "dim" in
                :attr:`hparams`. The input embedding is typically an
                aggregation of word embedding and position embedding.
            sequence_length: A 1D :tensor:`LongTensor` of shape
                ``[batch_size]``. Input tokens beyond respective sequence
                lengths are masked out automatically.

        Returns:
            A Tensor of shape ``[batch_size, max_time, dim]`` containing the
            encoded vectors.
        """
        # Multiply input embedding with the sqrt of its dimension for
        # normalization

        inputs_padding = 1 - sequence_mask(
            sequence_length, inputs.size()[1]).float()
        if self._hparams.use_bert_config:
            ignore_padding = attn.attention_bias_ignore_padding(
                inputs_padding, bias_value=-1e4)
        else:
            ignore_padding = attn.attention_bias_ignore_padding(
                inputs_padding)
        encoder_self_attention_bias = ignore_padding

        input_embedding = inputs
        if self._hparams.use_bert_config:
            x = self.input_normalizer(input_embedding)
            x = self.embed_dropout(x)
        else:
            x = self.embed_dropout(input_embedding)

        for i in range(self._hparams.num_blocks):
            # trivial difference between BERT and original Transformer
            if self._hparams.use_bert_config:
                _queries_input = x
            else:
                _queries_input = self.self_attn_layer_norm[i](x)

            attention_output = self.self_attns[i](
                queries=_queries_input,
                memory=_queries_input,
                memory_attention_bias=encoder_self_attention_bias,
            )

            attention_output = self.residual_dropout(attention_output)

            x = x + attention_output

            poswise_network = self.poswise_networks[i]
            poswise_normalizer = self.poswise_layer_norm[i]

            if self._hparams.use_bert_config:
                x = poswise_normalizer(x)
                y = x
            else:
                y = poswise_normalizer(x)

            original_shape = y.size()

            y = y.view(-1, self._hparams.dim)

            layer_output = poswise_network(y)
            sub_output = self.residual_dropout(layer_output)
            sub_output = sub_output.view(original_shape)

            x = x + sub_output
            if self._hparams.use_bert_config:
                x = self.output_layer_norm[i](x)

        if not self._hparams.use_bert_config:
            x = self.final_layer_norm(x)
        return x

    @property
    def output_size(self) -> int:
        return self._hparams.dim



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for vocabulary related operations.
"""
import tempfile
import unittest

from texar.torch.data import vocabulary


class VocabularyTest(unittest.TestCase):
    """Tests vocabulary related operations.
    """

    def test_make_defaultdict(self):
        """Tests the _make_defaultdict function.
        """
        keys = ['word', '词']
        values = [0, 1]
        default_value = -1

        dict_ = vocabulary._make_defaultdict(keys, values, default_value)

        self.assertEqual(len(dict_), 2)
        self.assertEqual(dict_['word'], 0)
        self.assertEqual(dict_['词'], 1)
        self.assertEqual(dict_['sth_else'], -1)

    def test_vocab_construction(self):
        """Test vocabulary construction.
        """
        vocab_list = ['word', '词']
        vocab_file = tempfile.NamedTemporaryFile()
        vocab_file.write('\n'.join(vocab_list).encode("utf-8"))
        vocab_file.flush()

        vocab = vocabulary.Vocab(vocab_file.name)

        self.assertEqual(vocab.size, len(vocab_list) + 4)
        self.assertEqual(
            set(vocab.token_to_id_map_py.keys()),
            set(['word', '词'] + vocab.special_tokens))

        # Tests UNK token
        unk_token_id = vocab.map_tokens_to_ids_py(['new'])
        unk_token_text = vocab.map_ids_to_tokens_py(unk_token_id)
        self.assertEqual(unk_token_text[0], vocab.unk_token)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for pre-trained BERT tokenizer utils.
"""

import unittest

from texar.torch.data.tokenizers.bert_tokenizer_utils import (
    BasicTokenizer, WordpieceTokenizer, _is_control,
    _is_punctuation, _is_whitespace)


class BERTTokenizerUtilsTest(unittest.TestCase):

    def test_chinese(self):

        tokenizer = BasicTokenizer()

        self.assertListEqual(
            tokenizer.tokenize(u"ah\u535A\u63A8zz"),
            [u"ah", u"\u535A", u"\u63A8", u"zz"])

    def test_basic_tokenizer_lower(self):

        tokenizer = BasicTokenizer(do_lower_case=True)

        self.assertListEqual(
            tokenizer.tokenize(u" \tHeLLo!how  \n Are yoU?  "),
            ["hello", "!", "how", "are", "you", "?"])
        self.assertListEqual(tokenizer.tokenize(u"H\u00E9llo"), ["hello"])

    def test_basic_tokenizer_no_lower(self):

        tokenizer = BasicTokenizer(do_lower_case=False)

        self.assertListEqual(
            tokenizer.tokenize(u" \tHeLLo!how  \n Are yoU?  "),
            ["HeLLo", "!", "how", "Are", "yoU", "?"])

    def test_wordpiece_tokenizer(self):

        vocab_tokens = [
            "[UNK]", "[CLS]", "[SEP]", "want", "##want", "##ed", "wa", "un", "runn",
            "##ing"
        ]

        vocab = {}
        for (i, token) in enumerate(vocab_tokens):
            vocab[token] = i
        tokenizer = WordpieceTokenizer(vocab=vocab, unk_token="[UNK]")

        self.assertListEqual(tokenizer.tokenize(""), [])

        self.assertListEqual(
            tokenizer.tokenize("unwanted running"),
            ["un", "##want", "##ed", "runn", "##ing"])

        self.assertListEqual(
            tokenizer.tokenize("unwantedX running"), ["[UNK]", "runn", "##ing"])

    def test_is_whitespace(self):

        self.assertTrue(_is_whitespace(u" "))
        self.assertTrue(_is_whitespace(u"\t"))
        self.assertTrue(_is_whitespace(u"\r"))
        self.assertTrue(_is_whitespace(u"\n"))
        self.assertTrue(_is_whitespace(u"\u00A0"))

        self.assertFalse(_is_whitespace(u"A"))
        self.assertFalse(_is_whitespace(u"-"))

    def test_is_control(self):

        self.assertTrue(_is_control(u"\u0005"))

        self.assertFalse(_is_control(u"A"))
        self.assertFalse(_is_control(u" "))
        self.assertFalse(_is_control(u"\t"))
        self.assertFalse(_is_control(u"\r"))

    def test_is_punctuation(self):

        self.assertTrue(_is_punctuation(u"-"))
        self.assertTrue(_is_punctuation(u"$"))
        self.assertTrue(_is_punctuation(u"`"))
        self.assertTrue(_is_punctuation(u"."))

        self.assertFalse(_is_punctuation(u"A"))
        self.assertFalse(_is_punctuation(u" "))


if __name__ == "__main__":
    unittest.main()
